from bfieldtools import mesh_magnetics
from bfieldtools.utils import load_example_mesh

import pytest

import numpy as np
from numpy.testing import assert_allclose


def test_magnetic_field_coupling():

    mesh = load_example_mesh("unit_disc")

    points = mesh.vertices + np.array([0, 0, 20])

    B_1_chunk = mesh_magnetics.magnetic_field_coupling(mesh, points, Nchunks=1)
    B_2_chunk = mesh_magnetics.magnetic_field_coupling(mesh, points, Nchunks=2)

    assert_allclose(B_1_chunk, B_2_chunk, atol=1e-16)

    B_q_1 = mesh_magnetics.magnetic_field_coupling(mesh, points, quad_degree=1)
    B_q_2 = mesh_magnetics.magnetic_field_coupling(mesh, points, quad_degree=2)
    B_q_3 = mesh_magnetics.magnetic_field_coupling(mesh, points, quad_degree=3)

    assert_allclose(B_q_1, B_q_2, atol=1e-3 * np.mean(np.abs(B_q_2)))
    assert_allclose(B_q_2, B_q_3, atol=1e-3 * np.mean(np.abs(B_q_2)))

    B_a = mesh_magnetics.magnetic_field_coupling(mesh, points, analytic=True)

    assert_allclose(B_a, B_q_1, atol=1e-3 * np.mean(np.abs(B_a)))
    assert_allclose(B_a, B_q_2, atol=1e-3 * np.mean(np.abs(B_a)))
    assert_allclose(B_a, B_q_3, atol=1e-3 * np.mean(np.abs(B_a)))


def test_magnetic_scalar_potential_coupling():

    mesh = load_example_mesh("unit_disc")

    points = mesh.vertices + np.array([0, 0, 20])

    U_1_chunk = mesh_magnetics.scalar_potential_coupling(mesh, points, Nchunks=1)
    U_2_chunk = mesh_magnetics.scalar_potential_coupling(mesh, points, Nchunks=2)

    assert_allclose(U_1_chunk, U_2_chunk, atol=1e-16)

    U_exact = mesh_magnetics.scalar_potential_coupling(mesh, points, approx_far=False)
    U_approx_far = mesh_magnetics.scalar_potential_coupling(
        mesh, points, approx_far=True, margin=0.5
    )

    assert_allclose(U_exact, U_approx_far, atol=2e-3 * np.mean(np.abs(U_exact)))


def test_magnetic_vector_potential_coupling():

    mesh = load_example_mesh("unit_disc")

    points = mesh.vertices + np.array([0, 0, 20])

    A_1_chunk = mesh_magnetics.vector_potential_coupling(mesh, points, Nchunks=1)
    A_2_chunk = mesh_magnetics.vector_potential_coupling(mesh, points, Nchunks=2)

    assert_allclose(A_1_chunk, A_2_chunk, atol=1e-16)

    A_exact = mesh_magnetics.vector_potential_coupling(mesh, points, approx_far=False)
    A_approx_far = mesh_magnetics.vector_potential_coupling(
        mesh, points, approx_far=True, margin=0.5
    )
    A_approx_far_clusters = mesh_magnetics.vector_potential_coupling(
        mesh, points, approx_far=True, margin=0.5, chunk_clusters=True
    )

    assert_allclose(A_exact, A_approx_far, atol=2e-3 * np.mean(np.abs(A_exact)))
    assert_allclose(
        A_exact, A_approx_far_clusters, atol=2e-3 * np.mean(np.abs(A_exact))
    )



"""

Analytic integral for vectorized field / potential computation

"""

__all__ = [
    "c_coeffs",
    "d_distance",
    "gamma0",
    "omega",
    "potential_dipoles",
    "potential_vertex_dipoles",
    "triangle_potential_approx",
    "triangle_potential_dipole_linear",
    "triangle_potential_uniform",
    "x_distance",
    "x_distance2",
]

import numpy as np


def determinant(a):
    """ Faster determinant for the two last dimensions of 'a'
    """
    det = a[..., 0, 0] * (a[..., 1, 1] * a[..., 2, 2] - a[..., 2, 1] * a[..., 1, 2])
    det += a[..., 0, 1] * (a[..., 1, 2] * a[..., 2, 0] - a[..., 2, 2] * a[..., 1, 0])
    det += a[..., 0, 2] * (a[..., 1, 0] * a[..., 2, 1] - a[..., 2, 0] * a[..., 1, 1])
    return det


def norm(vecs):
    """ Faster vector norm for the last dimension of 'vecs'
    """
    return np.sqrt(np.einsum("...i,...i", vecs, vecs))


def cross(r1, r2):
    """ Cross product without overhead for the last dimensions of 'r1' and 'r2'
    """
    result = np.zeros(r1.shape)
    result[..., 0] = r1[..., 1] * r2[..., 2] - r1[..., 2] * r2[..., 1]
    result[..., 1] = r1[..., 2] * r2[..., 0] - r1[..., 0] * r2[..., 2]
    result[..., 2] = r1[..., 0] * r2[..., 1] - r1[..., 1] * r2[..., 0]
    return result


def gamma0(R, reg=1e-13, symmetrize=True):
    """ 1/r integrals over the edges of a triangle called gamma_0
        (line charge potentials).

        **NOTE: MAY NOT BE VERY PRECISE FOR POINTS DIRECTLY AT TRIANGLE
        EDGES.**

        Parameters
        ----------
        R : ndarray (..., N_triverts, xyz)
            displacement vectors (r-r') between Neval evaluation points (r)
            and the 3 vertices of the Ntri triangles/triangle.
        reg: float, a small value added to the arguments of the logarithm,
             regularizes the values very close to the line segments
        symmetrize: recalculates the result for by mirroring
                    the evaluation points with respect the line segment
                    mid point to get rid off the badly behaving points on the
                    negative extension of the line segment


        Returns
        -------
        res: array (Neval, Nverts)
            The analytic integrals for each vertex/edge

    """
    edges = np.roll(R[0], 2, -2) - np.roll(R[0], 1, -2)
    dotprods1 = np.einsum("...i,...i", np.roll(R, 1, -2), edges)
    dotprods2 = np.einsum("...i,...i", np.roll(R, 2, -2), edges)
    en = norm(edges)
    del edges
    n = norm(R)
    # Regularize s.t. neither the denominator or the numerator can be zero
    # Avoid numerical issues directly at the edge
    nn1 = np.roll(n, 1, -1) * en
    nn2 = np.roll(n, 2, -1) * en
    res = np.log((nn1 + dotprods1 + reg) / (nn2 + dotprods2 + reg))

    # Symmetrize the result since on the negative extension of the edge
    # there's division of two small values resulting numerical instabilities
    # (also incompatible with adding the reg value)
    if symmetrize:
        mask = ((np.abs(dotprods1 + nn1)) < 1e-12) * (dotprods1 + dotprods2 < 0)
        res[mask] = -np.log(
            (nn1[mask] - dotprods1[mask] + reg) / (nn2[mask] - dotprods2[mask] + reg)
        )

    res /= en
    return -res


def omega(R):
    """ Calculate the solid angle of a triangles

        see
        A. Van Oosterom and J. Strackee
        IEEE TRANSACTIONS ON BIOMEDICAL ENGINEERING,
        VOL. BME-30, NO. 2, 1983

        Parameters
        ----------
        R : ndarray (Neval, (Ntri), N_triverts, xyz)
            displacement vectors (r-r') of Ntri triangles
            and Neval evaluation points for the 3 vertices
            of the triangles/triangle.

            The shape of R can any with the constraint that
            the last dimenion corrsponds to coordinates (x, y, z) and the
            second last dimension to triangle vertices (vert1, vert2, vert3)

        Returns
        -------
        sa: (Neval, (Ntri))
            Solid angles of subtened by triangles at evaluation points
    """
    # Distances
    d = norm(R)
    # Scalar triple products
    stp = determinant(R)
    # Denominator
    denom = np.prod(d, axis=-1)
    for i in range(3):
        j = (i + 1) % 3
        k = (i + 2) % 3
        #        denom += np.sum(R[..., i, :]*R[..., j, :], axis=-1)*d[..., k]
        denom += np.einsum("...i,...i,...", R[..., i, :], R[..., j, :], d[..., k])
    # Solid angles
    sa = -2 * np.arctan2(stp, denom)
    return sa


def x_distance(R, tn, ta=None):
    """ Signed distances in the triangle planes from the opposite
        edge towards the node for all evaluation points in R

        The distances are normalized to one at the node if areas are given
        The distances are multiplied by the edge lenght if areass are None

        Parameters:

            R: ndarray (... Ntri, Nverts, xyz)
                displacement vectors (coordinates)
            tn: ndarray (Ntri, 3)
                triangle normals
            ta: ndarray (Ntri)
                triangle areas
                if None, normalizization with double area is not carried out

        returns:
            ndaarray (..., Ntri, N_triverts (3)), distance in the triangle plane
    """
    edges = np.roll(R[0], 2, -2) - np.roll(R[0], 1, -2)
    if ta is not None:
        edges /= 2 * ta[:, None, None]
    edges = -cross(edges, tn[:, None, :])
    return np.einsum("...k,...k->...", np.roll(R, 1, -2), edges)


def x_distance2(mesh):
    """ Signed distances in the triangle planes from the opposite
        edge towards the node for all evalution points in R
    """
    # TODO: with gradient, needs mesh info
    pass


def d_distance(R, tn):
    """ Signed distance from the triangle plane for each triangle

        Parameters:

            R: ndarray (... Ntri, Nverts, xyz)
                displacement vectors (coordinates)
            tn: ndarray (Ntri, 3)
                triangle normals

        Returns:

            ndarray (..., Ntri, N_triverts (3)) of signed distances
    """
    return np.einsum("...ki,ki->...k", np.take(R, 0, -2), tn)


def c_coeffs(R, ta):
    """ Cotan-coeffs

        Parameters:

            R: ndarray (... Ntri, Nverts, xyz)
                displacement vectors (coordinates)
            ta: ndarray (Ntri)
                triangle areas


        Returns:

            ndarray (..., Ntri, N_triverts (3))
    """
    edges = np.roll(R[0], 2, -2) - np.roll(R[0], 1, -2)
    return np.einsum("...ik,...jk->...ij", edges, edges / (2 * ta[:, None, None]))


def triangle_potential_uniform(R, tn, planar=False):
    """ 1/r potential of a uniform triangle

        for original derivation see
        A. S. Ferguson, Xu Zhang and G. Stroink,
        "A complete linear discretization for calculating the magnetic field
        using the boundary element method,"
        in IEEE Transactions on Biomedical Engineering,
        vol. 41, no. 5, pp. 455-460, May 1994.
        doi: 10.1109/10.293220


        Parameters
        ----------

        R : (Neval, (Ntri), 3, 3) array
            Displacement vectors (Neval, (Ntri), Ntri_verts, xyz)
        tn : ((Ntri), 3) array
            Triangle normals (Ntri, dir)
        planar: boolean
            If True, assume all the triangles and the evaluation points
                    are on the same plane (for speed), leaves out the
                    omega term

        Returns
        -------
        result: result:  ndarray (Neval, (Ntri))
            Resultant 1/r potential for each triangle (Ntri)
            at the field evaluation points (Neval)

    """
    x = x_distance(R, tn, None)
    result = np.einsum("...i,...i", gamma0(R), x)
    if not planar:
        result += d_distance(R, tn) * omega(R)
    return result


def triangle_potential_approx(Rcenters, ta, reg=1e-12):
    """ 1/r potential of a uniform triangle using centroid approximation

        Calculates 1/R potentials for triangle centroids
        (The singularity at the centroid is handled with the very small
        reg value, but anyway the values close to the centroid are inexact)

        Parameters
        ----------
        Rcenters : (N, (Ntri), 3) array
            Displacement vectors (Neval, Ntri, xyz)
            from triangle centers
        ta : (Ntri) array
            Triangle areas

        reg: float
            Regularization value used in approximation

        Returns
        -------
        result: result:  ndarray (...., Ntri, Ntri_verts)
            Resultant 1/r potential for each node (Ntri_verts)
            in each triangle (Ntri) in the displacement vectors R

    """
    result = ta / (norm(Rcenters) + reg)
    return result


def potential_dipoles(R, face_normals, face_areas):
    """ Approximate the potential of linearly varying dipole density by
        by dipoles at each face

    Parameters
            R : ndarray (Neval, Ntri, Ntri_verts, N_xyz)
                Displacement vectors
            face_normals: ndarray (Ntri, 3)
                normals for each triangle
            face_areas: ndarray (Ntri,)
                areas for each triangle

    Return
        Potential approximation for vertex in each face
        pot: (Neval, Ntri, Ntriverts)
    """
    nn = face_normals
    # Calculate quadrature points corresponding to linear shape functions (Ok?)
    weights = np.array([[0.5, 0.25, 0.25], [0.25, 0.5, 0.25], [0.25, 0.25, 0.5]])
    #    weights = np.eye(3)
    #    weights = np.ones((3,3))/3
    # Combine vertices for quadrature points
    Rquad = np.einsum("...ij,ik->...kj", R, weights)
    pot = np.einsum("ik, ...ijk->...ij", nn, Rquad) / (norm(Rquad) ** 3)
    pot = pot * (face_areas[:, None] / 3)

    return pot


def potential_vertex_dipoles(R, vertex_normals, vertex_areas):
    """ Approximate the potential of linearly varying dipole density by
        by dipoles at each vertex

    Parameters
            R : ndarray (Neval, Nvertex, N_xyz)
                Displacement vectors
            vertex_normals: ndarray (Nvertex, 3)
                normals for each triangle
            vertex_areas: ndarray (Nvertex,)
                areas for each triangle

    Return
        Potential approximation for vertex in each face
        pot: (Neval, Ntri, Ntriverts)
    """
    nn = vertex_normals
    pot = np.einsum("ik, lik->li", nn, R) / (norm(R) ** 3)
    pot *= vertex_areas

    return pot


def triangle_potential_dipole_linear(R, tn, ta):
    """ Potential of dipolar density with magnitude of a
        linear shape function on a triangle, "omega_i" in de Munck's paper

        for the original derivation, see:
        J. C. de Munck, "A linear discretization of the volume mesh_conductor
        boundary integral equation using analytically integrated elements
        (electrophysiology application),"
        in IEEE Transactions on Biomedical Engineering,
        vol. 39, no. 9, pp. 986-990, Sept. 1992.
        doi: 10.1109/10.256433

        Parameters
        ----------

        R : (..., Ntri, 3, 3) array
            Displacement vectors (...., Ntri, Ntri_verts, xyz)
        tn : ((Ntri), 3) array
            Triangle normals (Ntri, dir)
        ta : (Ntri), array
            Triangle areas (Ntri, dir)

        Returns
        -------
        result:  ndarray (...., Ntri, Ntri_verts)
            Resultant dipolar potential for each shape functions (Ntri_verts)
            in each triangle (Ntri) at the points
            corresponding to displacement vectors in R

    """

    result = np.einsum(
        "...i,...ij,...->...j",
        gamma0(R),
        c_coeffs(R, ta),
        d_distance(R, tn),
        optimize=True,
    )
    x_dists = x_distance(R, tn, ta)
    result -= x_dists * omega(R)[..., :, None]

    return result



from bfieldtools import sphtools

import pytest


import numpy as np

from numpy.testing import assert_allclose


def test_coord_changes():

    point = np.array([[0.4, 0.4, 1]])

    sphpoint = sphtools.cartesian2spherical(point)

    point2 = sphtools.spherical2cartesian(sphpoint)

    assert_allclose(point, point2)


def test_sph_eval():
    """
    Simply test that the function evaluate, assert shapes but nothing more
    """

    l = 4

    x = np.array([0.4, 0.4, 1])
    theta = np.array([0.4, 0.4, 1])
    phi = np.array([0.4, 0.4, 1])

    for m in [-2, 0, 2]:
        m = 2

        assert sphtools.lpmn_em(l, m, x).shape == (3,)
        assert sphtools.derlpmn_em(l, m, x).shape == (3,)

        assert sphtools.xlm(l, m, theta).shape == (3,)

        assert sphtools.ylm(l, m, theta, phi).shape == (3,)

        assert sphtools.derxlm(l, m, theta).shape == (3,)

        assert sphtools.sinxlm(l, m, theta).shape == (3,)

        assert sphtools.dthylm(l, m, theta, phi).shape == (3,)

        assert sphtools.dphiylm(l, m, theta, phi).shape == (3,)

        assert sphtools.Plm(l, m, theta, phi).shape == (3, 3)

        assert sphtools.Blm(l, m, theta, phi).shape == (3, 3)

        assert sphtools.Wlm(l, m, theta, phi).shape == (3, 3)
        assert sphtools.Vlm(l, m, theta, phi).shape == (3, 3)


def test_potential():

    p = np.array([[1, 0, 0], [-1, 0, 0]])

    acoeffs = np.array([1, 0, 0])
    bcoeffs = np.array([0, 0, 0])
    lmax = 1

    U = sphtools.potential(p, acoeffs, bcoeffs, lmax)

    assert U[0] == -U[-1]


def test_field():

    p = np.array([[1, 0, 0], [-1, 2, 0]])

    # make a homogeneous field
    acoeffs = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    bcoeffs = np.array([1, 0, 0, 0, 0, 0, 0, 0])
    lmax = 2

    B = sphtools.field(p, acoeffs, bcoeffs, lmax)

    assert_allclose(B[0], B[1], atol=1e-16)

    basis = sphtools.basis_fields(p, lmax)

    B_b = basis[0] @ acoeffs + basis[1] @ bcoeffs

    assert_allclose(B_b[0], B_b[1], atol=1e-16)

    assert_allclose(B, B_b, atol=1e-16)

    p = np.array([[0, 1, 0], [100, 2, 500]])

    # make a homogeneous field
    acoeffs = np.array([0, 0, 0, 0, 0, 0, 0, 0])
    bcoeffs = np.array([0, 1, 0, 0, 0, 0, 0, 0])
    lmax = 2

    B = sphtools.field(p, acoeffs, bcoeffs, lmax)

    assert_allclose(B[0], B[1], atol=1e-16)

    basis = sphtools.basis_fields(p, lmax)

    B_b = basis[0] @ acoeffs + basis[1] @ bcoeffs

    assert_allclose(B_b[0], B_b[1], atol=1e-16)

    assert_allclose(B, B_b, atol=1e-16)


def test_innerproduct():

    sph = sphtools.SphBasis(40)

    atol = 1e-12

    Ylm1 = sphtools.ylm(2, 1, sph.sqp[:, 1], sph.sqp[:, 2])
    Ylm2 = sphtools.ylm(2, 2, sph.sqp[:, 1], sph.sqp[:, 2])

    assert_allclose(sph.innerproduct(Ylm1[:, None], Ylm1[:, None]), 1, atol=atol)
    assert_allclose(sph.innerproduct(Ylm2[:, None], Ylm2[:, None]), 1, atol=atol)
    assert_allclose(sph.innerproduct(Ylm1[:, None], Ylm2[:, None]), 0, atol=atol)

    Vlm1 = sphtools.Vlm(2, 1, sph.sqp[:, 1], sph.sqp[:, 2])
    Vlm2 = sphtools.Vlm(5, 2, sph.sqp[:, 1], sph.sqp[:, 2])

    assert_allclose(sph.innerproduct(Vlm1, Vlm1), 1, atol=atol)
    assert_allclose(sph.innerproduct(Vlm2, Vlm2), 1, atol=atol)
    assert_allclose(sph.innerproduct(Vlm1, Vlm2), 0, atol=atol)

    Wlm1 = sphtools.Wlm(3, 1, sph.sqp[:, 1], sph.sqp[:, 2])
    Wlm2 = sphtools.Wlm(5, 2, sph.sqp[:, 1], sph.sqp[:, 2])

    assert_allclose(sph.innerproduct(Wlm1, Wlm1), 1, atol=atol)
    assert_allclose(sph.innerproduct(Wlm2, Wlm2), 1, atol=atol)
    assert_allclose(sph.innerproduct(Wlm1, Wlm2), 0, atol=atol)

    Xlm1 = sphtools.Xlm(3, 1, sph.sqp[:, 1], sph.sqp[:, 2])
    Xlm2 = sphtools.Xlm(4, 2, sph.sqp[:, 1], sph.sqp[:, 2])

    assert_allclose(sph.innerproduct(Xlm1, Xlm1), 1, atol=atol)
    assert_allclose(sph.innerproduct(Xlm2, Xlm2), 1, atol=atol)
    assert_allclose(sph.innerproduct(Xlm1, Xlm2), 0, atol=atol)

    assert_allclose(sph.innerproduct(Xlm1, Vlm1), 0, atol=atol)
    assert_allclose(sph.innerproduct(Xlm1, Vlm2), 0, atol=atol)
    assert_allclose(sph.innerproduct(Xlm2, Vlm1), 0, atol=atol)
    assert_allclose(sph.innerproduct(Xlm2, Vlm2), 0, atol=atol)

    assert_allclose(sph.innerproduct(Wlm1, Vlm1), 0, atol=atol)
    assert_allclose(sph.innerproduct(Wlm1, Vlm2), 0, atol=atol)
    assert_allclose(sph.innerproduct(Wlm2, Vlm1), 0, atol=atol)
    assert_allclose(sph.innerproduct(Wlm2, Vlm2), 0, atol=atol)


def test_mesh_coupling():
    """
        Test compute_sphcoeffs_mesh with sphere
    """

    from bfieldtools.sphtools import compute_sphcoeffs_mesh
    from bfieldtools.utils import load_example_mesh
    from bfieldtools.sphtools import basis_potentials, basis_fields
    from bfieldtools.sphtools import ylm, cartesian2spherical
    from bfieldtools.mesh_calculus import mass_matrix

    mesh = load_example_mesh("unit_sphere")
    mesh.vertices *= 1 / np.sqrt(mesh.area / (4 * np.pi))  # Scale to unit radius
    R = 2
    mesh.vertices *= R  # Scale to R

    c = compute_sphcoeffs_mesh(mesh, 3)

    # Test potential
    sp = cartesian2spherical(mesh.vertices)
    M = mass_matrix(mesh, lumped=True)
    u1, u2 = basis_potentials(mesh.vertices, 3)
    diff1 = []
    diff2 = []

    for ll in range(1, 4):
        for m in range(-ll, ll + 1):
            s = ylm(ll, m, sp[:, 1], sp[:, 2])
            p = u1 @ c[0] @ s
            # p should be p= ll/(2*ll+1)s, test this
            coeff = s @ M @ p / (s @ M @ s)
            diff1.append(coeff - ll / (2 * ll + 1))
            p = u2 @ c[1] @ s
            # p should be p= -(ll+1)/(2*ll+1)s, test this
            coeff = s @ M @ p / (s @ M @ s)
            diff2.append(coeff + (ll + 1) / (2 * ll + 1))

    # The integration accuracy is quite low so set the tolerance high
    assert np.allclose(diff1, 0, atol=1e-2)
    assert np.allclose(diff2, 0, atol=1e-2)

    # Test field
    b1, b2 = basis_fields(mesh.vertices, 3)
    b1 = np.einsum("ijk,ij->ik", b1, mesh.vertex_normals)
    b2 = np.einsum("ijk,ij->ik", b2, mesh.vertex_normals)
    diff1 = []
    diff2 = []
    mu0 = 4 * np.pi * 1e-7
    for ll in range(1, 4):
        for m in range(-ll, ll + 1):
            s = ylm(ll, m, sp[:, 1], sp[:, 2])
            p = b1 @ c[0] @ s
            # p should be p= mu0*(ll+1)*ll/(2*ll+1)/R s, test this
            coeff = s @ M @ p / (s @ M @ s)
            diff1.append(coeff / mu0 - (ll + 1) * ll / (2 * ll + 1) / R)
            p = b2 @ c[1] @ s
            # p should be p= mu0*(ll+1)*ll/(2*ll+1)/R s, test this
            coeff = s @ M @ p / (s @ M @ s)
            diff2.append(coeff / mu0 - (ll + 1) * ll / (2 * ll + 1) / R)

    assert np.allclose(diff1, 0, atol=1e-2)
    assert np.allclose(diff2, 0, atol=1e-2)



"""
Field interpolation example using equivalent surface currents
==============================================================

"""
import numpy as np
from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from mayavi import mlab
import trimesh

import mne

PLOT = True
SAVE_FIGURES = True
IMPORT_MNE_DATA = True

SAVE_MNE_DATA = True
SAVE_DIR = "./MNE interpolation/"


from pyface.api import GUI

_gui = GUI()


#%%
# First, let's import the MEG data

if IMPORT_MNE_DATA:

    from mne.datasets import sample

    data_path = sample.data_path()
    fname = data_path + "/MEG/sample/sample_audvis-ave.fif"
    # Reading
    condition = "Left Auditory"
    evoked = mne.read_evokeds(fname, condition=condition, baseline=(None, 0), proj=True)
    evoked.pick_types(meg="mag")
    # evoked.plot(exclude=[], time_unit="s")

    i0, i1 = evoked.time_as_index(0.08)[0], evoked.time_as_index(0.09)[0]
    field = evoked.data[:, i0:i1].mean(axis=1)

    # Read BEM for surface geometry and transform to correct coordinate system
    import os.path as op

    subject = "sample"
    subjects_dir = op.join(data_path, "subjects")
    bem_fname = op.join(
        subjects_dir, subject, "bem", subject + "-5120-5120-5120-bem-sol.fif"
    )
    bem = mne.read_bem_solution(bem_fname)

    # Head mesh 0
    # Innerskull mesh 2
    surf_index = 2

    trans_fname = op.join(data_path, "MEG", "sample", "sample_audvis_raw-trans.fif")
    trans0 = mne.read_trans(trans_fname)
    R = trans0["trans"][:3, :3]
    t = trans0["trans"][:3, 3]
    # Surface from MRI to HEAD
    rr = (bem["surfs"][surf_index]["rr"] - t) @ R
    # Surface from HEAD to DEVICE
    trans1 = evoked.info["dev_head_t"]
    R = trans1["trans"][:3, :3]
    t = trans1["trans"][:3, 3]
    rr = (rr - t) @ R

    innerskull = trimesh.Trimesh(rr, bem["surfs"][surf_index]["tris"])

    surf_index = 0

    R = trans0["trans"][:3, :3]
    t = trans0["trans"][:3, 3]
    # Surface from MRI to HEAD
    rr = (bem["surfs"][surf_index]["rr"] - t) @ R
    # Surface from HEAD to DEVICE
    R = trans1["trans"][:3, :3]
    t = trans1["trans"][:3, 3]
    rr = (rr - t) @ R
    head = trimesh.Trimesh(rr, bem["surfs"][surf_index]["tris"])

    mesh = head

    # Sensor locations and directions in DEVICE coordinate system
    p = np.array(
        [
            ch["loc"][:3]
            for ch in evoked.info["chs"]
            if ch["ch_name"][-1] == "1" and ch["ch_name"][:3] == "MEG"
        ]
    )
    n = np.array(
        [
            ch["loc"][-3:]
            for ch in evoked.info["chs"]
            if ch["ch_name"][-1] == "1" and ch["ch_name"][:3] == "MEG"
        ]
    )

    if PLOT:
        # Plot sensor locations and directions
        fig = mlab.figure(bgcolor=(1, 1, 1))
        mlab.triangular_mesh(*innerskull.vertices.T, innerskull.faces)
        mlab.triangular_mesh(
            *head.vertices.T, head.faces, color=(0.5, 0.5, 0.5), opacity=0.5
        )
        mlab.quiver3d(*p.T, *n.T, mode="arrow")

        fig.scene.isometric_view()

        if SAVE_FIGURES:
            mlab.savefig(SAVE_DIR + "MEG_geometry.png", magnification=4, figure=fig)

    if SAVE_MNE_DATA:
        np.savez(
            SAVE_DIR + "mne_data.npz",
            mesh=head,
            p=p,
            n=n,
            vertices=head.vertices,
            faces=head.faces,
        )
        evoked.save(SAVE_DIR + "left_auditory-ave.fif")


else:

    with np.load(SAVE_DIR + "mne_data.npz", allow_pickle=True) as data:
        mesh = data["mesh"]
        p = data["p"]
        n = data["n"]
        mesh = trimesh.Trimesh(vertices=data["vertices"], faces=data["faces"])

    evoked = mne.Evoked(SAVE_DIR + "left_auditory-ave.fif")


#%%
# Fit the surface current for the auditory evoked response


c = MeshConductor(mesh_obj=mesh, basis_name="suh", N_suh=150)
M = c.mass

sensor_coupling = np.einsum("ijk,ij->ik", c.B_coupling(p), n)
# a = np.linalg.pinv(sensor_coupling, rcond=1e-15) @ field
ss = np.linalg.svd(sensor_coupling @ sensor_coupling.T, False, False)

# reg_exps = [0.5, 1, 2, 3, 4, 5, 6, 7, 8]
reg_exps = [1]
rel_errors = []

for reg_exp in reg_exps:
    _lambda = np.max(ss) * (10 ** (-reg_exp))
    # Laplacian in the suh basis is diagonal
    BB = sensor_coupling.T @ sensor_coupling + _lambda * (-c.laplacian) / np.max(
        abs(c.laplacian)
    )
    a = np.linalg.solve(BB, sensor_coupling.T @ field)
    s = StreamFunction(a, c)
    b_filt = sensor_coupling @ s

    rel_error = np.linalg.norm(b_filt - field) / np.linalg.norm(field)
    print("Relative error:", rel_error * 100, "%")
    rel_errors.append(rel_error)

    if PLOT:
        fig = mlab.figure(bgcolor=(1, 1, 1))
        surf = s.plot(False, figure=fig)
        surf.actor.mapper.interpolate_scalars_before_mapping = True
        surf.module_manager.scalar_lut_manager.number_of_colors = 16

        if SAVE_FIGURES:
            mlab.savefig(
                SAVE_DIR + "SUH_scalp_streamfunction.png", magnification=4, figure=fig
            )

#%%
# Interpolate MEG data to the sensor surface

from bfieldtools.utils import load_example_mesh

helmet = load_example_mesh("meg_helmet", process=False)
# Bring the surface roughly to the correct place
helmet.vertices[:, 2] -= 0.05

# Reset coupling by hand
c.B_coupling.reset()
mlab.figure(bgcolor=(1, 1, 1))
B_surf = np.sum(
    c.B_coupling(helmet.vertices) * helmet.vertex_normals[:, :, None], axis=1
)

if PLOT:
    fig = mlab.quiver3d(*p.T, *n.T, mode="arrow")
    scalars = B_surf @ s
    surf = mlab.triangular_mesh(
        *helmet.vertices.T, helmet.faces, scalars=scalars, colormap="seismic"
    )
    surf.actor.mapper.interpolate_scalars_before_mapping = True
    surf.module_manager.scalar_lut_manager.number_of_colors = 15

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "SUH_sensors_streamfunction.png", magnification=4, figure=fig
        )


#%%
# Calculate magnetic field in volumetric grid

Nvol = 30
x = np.linspace(-0.125, 0.125, Nvol)
vol_points = np.array(np.meshgrid(x, x, x, indexing="ij")).reshape(3, -1).T
# mlab.points3d(*vol_points.T)

c.B_coupling.reset()
Bvol_coupling = c.B_coupling(vol_points, Nchunks=100, analytic=True)
s = StreamFunction(a, c)
# s = StreamFunction(a, c)
Bvol = Bvol_coupling @ s

#%%
# Now, plot the computed magnetic field using streamlines

if PLOT:
    from bfieldtools.mesh_calculus import gradient

    fig = mlab.figure(bgcolor=(1, 1, 1))

    surf = s.plot(False, figure=fig)
    surf.actor.mapper.interpolate_scalars_before_mapping = True
    surf.module_manager.scalar_lut_manager.number_of_colors = 16

    vecs = mlab.pipeline.vector_field(
        *vol_points.T.reshape(3, Nvol, Nvol, Nvol), *Bvol.T.reshape(3, Nvol, Nvol, Nvol)
    )
    vecnorm = mlab.pipeline.extract_vector_norm(vecs)

    seed_points = mesh.vertices[mesh.faces].mean(axis=1) - 0.01 * mesh.face_normals

    seed_vals = c.basis @ c.inductance @ s
    seed_vals_grad = np.linalg.norm(gradient(seed_vals, c.mesh), axis=0)

    seed_vals = abs(seed_vals[mesh.faces].mean(axis=1)) ** 2
    seed_vals[seed_vals_grad > seed_vals_grad.max() / 1.8] = 0

    Npoints = 500
    seed_inds = np.random.choice(
        np.arange(len(seed_vals)), Npoints, False, seed_vals / seed_vals.sum()
    )

    seed_points = seed_points[seed_inds]

    streams = []

    for pi in seed_points:
        streamline = mlab.pipeline.streamline(
            vecnorm,
            integration_direction="both",
            colormap="BuGn",
            seed_visible=False,
            seedtype="point",
        )
        streamline.seed.widget.position = pi
        streamline.stream_tracer.terminal_speed = 3e-13
        streamline.stream_tracer.maximum_propagation = 0.1
        streamline.actor.property.render_lines_as_tubes = True
        streamline.actor.property.line_width = 4.0
        streams.append(streamline)

    # Custom colormap with alpha channel
    streamine = streams[0]
    lut = streamline.module_manager.scalar_lut_manager.lut.table.to_array()
    lut[:, -1] = np.linspace(0, 255, 256)
    streamline.module_manager.scalar_lut_manager.lut.table = lut
    streamline.module_manager.scalar_lut_manager.data_range = np.array(
        [1.0e-13, 1.0e-12]
    )

    for streamline in streams:
        streamline.stream_tracer.terminal_speed = 1e-13
        streamline.seed.widget.hot_spot_size = 0.1
        streamline.stream_tracer.initial_integration_step = 0.01
        streamline.stream_tracer.minimum_integration_step = 0.1

    sensors = mlab.quiver3d(*p.T, *n.T, mode="cylinder")
    sensors.glyph.glyph_source.glyph_source.height = 0.1
    sensors.actor.property.color = (0.5, 0.5, 0.5)
    sensors.actor.mapper.scalar_visibility = False
    sensors.glyph.glyph_source.glyph_source.resolution = 32
    sensors.glyph.glyph.scale_factor = 0.03
    # sensors.glyph.glyph_source.glyph_source.shaft_radius = 0.05

    fig.scene.camera.position = [
        0.637392177469018,
        0.07644693029292644,
        -0.07183513804689762,
    ]
    fig.scene.camera.focal_point = [
        -6.413459777832031e-05,
        0.01716560870409012,
        -0.0229007127850005,
    ]
    fig.scene.camera.view_angle = 30.0
    fig.scene.camera.view_up = [
        0.04390624852005244,
        0.3114421192517664,
        0.9492502555685007,
    ]
    fig.scene.camera.clipping_range = [0.3366362817578398, 1.0281065506557443]
    fig.scene.camera.compute_view_plane_normal()

    while fig.scene.light_manager is None:
        _gui.process_events()

    camera_light = fig.scene.light_manager.lights[0]
    camera_light.intensity = 0.7

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "SUH_streamlines_lateral.png", figure=fig, magnification=4
        )

    fig.scene.camera.position = [
        -6.413459777832031e-05,
        0.01716560870409012,
        0.6191735842078244,
    ]
    fig.scene.camera.focal_point = [
        -6.413459777832031e-05,
        0.01716560870409012,
        -0.0229007127850005,
    ]
    fig.scene.camera.view_angle = 30.0
    fig.scene.camera.view_up = [0.0, 1.0, 0.0]
    fig.scene.camera.clipping_range = [0.3381552363433513, 1.0261944997830243]
    fig.scene.camera.compute_view_plane_normal()

    if SAVE_FIGURES:
        mlab.savefig(
            SAVE_DIR + "SUH_streamlines_coronal.png", figure=fig, magnification=4
        )


#%%
# Finally, plot the surface current density itself

fig = mlab.figure(bgcolor=(1, 1, 1))
grad_s = gradient(c.basis @ s, mesh, rotated=True)
q = mlab.quiver3d(
    *(mesh.vertices[mesh.faces].mean(axis=1).T),
    *grad_s,
    colormap="viridis",
    mode="arrow"
)

mlab.triangular_mesh(*head.vertices.T, head.faces, color=(0.8, 0.8, 0.8), opacity=1.0)

fig.scene.camera.position = [
    0.4987072212753703,
    0.06469079487766746,
    -0.0014732384935239248,
]
fig.scene.camera.focal_point = [
    0.0018187984824180603,
    0.012344694641686624,
    -0.04367139294433087,
]
fig.scene.camera.view_angle = 30.0
fig.scene.camera.view_up = [
    -0.10720122151366927,
    0.23975383168819672,
    0.9648968848000314,
]
fig.scene.camera.clipping_range = [0.28329092545021717, 0.7772019991936254]
fig.scene.camera.compute_view_plane_normal()

if SAVE_FIGURES:
    mlab.savefig(
        SAVE_DIR + "SUH_surface_currents_lateral.png", figure=fig, magnification=4
    )



"""
Triangle dipole density approximation error
==================================================

Compare the exact solution for the potential of dipolar density with magnitude of a
linear shape function on a triangle with two approximations.

"""


from bfieldtools.integrals import (
    potential_vertex_dipoles,
    triangle_potential_dipole_linear,
    potential_dipoles,
)

from bfieldtools.mesh_calculus import mass_matrix
from bfieldtools.utils import load_example_mesh

import matplotlib.pyplot as plt


import numpy as np
from scipy.sparse import csc_matrix

mesh = load_example_mesh("unit_disc")


z = np.linspace(0, 10, 101)[1:]
x = np.zeros_like(z)
y = np.zeros_like(z)

points = np.array([x, y, z]).T

R = points[:, None, None, :] - mesh.vertices[mesh.faces][None, :, :, :]
R_v = points[:, None, :] - mesh.vertices[None, :, :]

vertex_areas = mass_matrix(mesh, lumped=True).diagonal()

approx_pot_v = potential_vertex_dipoles(R_v, mesh.vertex_normals, vertex_areas)

approx_pot_f = potential_dipoles(R, mesh.face_normals, mesh.area_faces)
exact_pot_f = triangle_potential_dipole_linear(R, mesh.face_normals, mesh.area_faces)

# Map faces -> vertices
Nf = len(mesh.faces)
Nv = len(mesh.vertices)
M0 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 0])), (Nf, Nv))
M1 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 1])), (Nf, Nv))
M2 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 2])), (Nf, Nv))
exact_pot_v = (
    exact_pot_f[:, :, 0] @ M0 + exact_pot_f[:, :, 1] @ M1 + exact_pot_f[:, :, 2] @ M2
)
approx_pot_fv = (
    approx_pot_f[:, :, 0] @ M0 + approx_pot_f[:, :, 1] @ M1 + approx_pot_f[:, :, 2] @ M2
)

mask = np.linalg.norm(mesh.vertices, axis=1) < 0.9
mask[5:] = False
RE_fv = (
    np.abs((approx_pot_fv[:, mask] - exact_pot_v[:, mask]) / exact_pot_v[1, mask]) * 100
)
RE_v = (
    np.abs((approx_pot_v[:, mask] - exact_pot_v[:, mask]) / exact_pot_v[1, mask]) * 100
)

plt.semilogy(z, approx_pot_fv[:, mask], "--", label="potential_dipoles")
plt.gca().set_prop_cycle(None)
plt.semilogy(z, approx_pot_v[:, mask], "-.", label="potential_vertex_dipoles")
plt.gca().set_prop_cycle(None)
plt.semilogy(z, exact_pot_v[:, mask], label="exact_vertex_dipoles")
plt.ylabel("Potential")
plt.xlabel("Distance")


plt.figure()
plt.semilogy(z, (approx_pot_fv - exact_pot_v)[:, mask], "--", label="potential_dipoles")
plt.gca().set_prop_cycle(None)
plt.semilogy(
    z, (approx_pot_v - exact_pot_v)[:, mask], "-.", label="potential_vertex_dipoles"
)
plt.ylabel("Error")
plt.xlabel("Distance")


plt.figure()
plt.semilogy(z, RE_fv, "--", label="potential_dipoles")
plt.gca().set_prop_cycle(None)
plt.semilogy(z, RE_v, "-.", label="potential_vertex_dipoles")
plt.ylabel("Relative error (%)")
plt.xlabel("Distance")



from bfieldtools import contour

import pytest

from numpy.testing import (
    assert_array_almost_equal,
    assert_array_equal,
    assert_allclose,
    assert_equal,
)


def test_scalar_contour():

    assert True


def test_simplify_contour():

    assert True



"""

Functions for creating contours (isolines) of a scalar function defined on a triangle mesh surface. Also contains functions for modifying the generated contours.

"""

__all__ = ["scalar_contour", "simplify_contour"]

import numpy as np

from scipy.sparse import eye as speye
from scipy.sparse import spdiags
from scipy.sparse.linalg import spsolve

from .mesh_calculus import gradient
from . import mesh_conductor


def scalar_contour(mesh, scalars, N_contours=10, contours=None, return_values=False):
    """
    Computes contour loops (isolines) for a scalar function defined on a mesh.
    The winding direction of the loops is defined according to the rotated gradient of the scalar

    Parameters
    ----------
    mesh: Trimesh object
        Trimesh object containing the mesh on which the scalar function is defined
    scalars: array-like or StreamFunction
        Vector containing the values of the scalar function at each of the mesh vertices.
        If StreamFunction, uses the vertex-wise values
    N_contours: int
        Number of contours to generate
    contours: array-like
        Optional argument for manual input of contour levels. Overrides `N_contours`
    return_values: Boolean
        If True, also return contour values

    Returns
    -------
    contour_polys: list
        list with length `N_contours`. Each list element is anumpy array containing the
        coordinats of each polygon vertex.
    contour_values: array-like
        Vector containing the scalar function value for each contour line,
        returned if return_values is True
    """

    if isinstance(scalars, mesh_conductor.StreamFunction):
        scalars = scalars.vert

    # Compute rotated gradient of scalars (e.g. current density from a stream function)
    g = gradient(scalars, mesh, rotated=True)

    if contours is None:
        # N evenly spaced contours that do not contain the point-like max and min contours
        contours = np.linspace(scalars.min(), scalars.max(), 2 * N_contours + 1)[1::2]

    edge_vals = scalars[mesh.edges_unique]
    contour_polys = []

    contour_values = []

    # Loop through
    for c in contours:

        # Get edges that contain the contour values
        edge_inds = (edge_vals.min(axis=1) <= c) * (edge_vals.max(axis=1) >= c)
        c_edge_vals = edge_vals[edge_inds]

        # Solve weights (barycentric coordinates) for each edge
        w0 = (c - c_edge_vals[:, 1]) / (c_edge_vals[:, 0] - c_edge_vals[:, 1])
        w1 = (-c + c_edge_vals[:, 0]) / (c_edge_vals[:, 0] - c_edge_vals[:, 1])

        # Calculate points linearly interpolated from vertices
        points = mesh.vertices[mesh.edges_unique[edge_inds]]
        points = points[:, 0] * w0[:, None] + points[:, 1] * w1[:, None]

        # Determine adjacency
        c_edges_in_faces = edge_inds[mesh.faces_unique_edges]
        c_faces = np.any(c_edges_in_faces, axis=1)
        edges_in_c_faces = mesh.faces_unique_edges[c_faces]

        if len(edges_in_c_faces) == 0:
            print("No contours at f=", c)
            continue

        # Each element of this array corresponds to a face containing the contour
        # The two values in the element are the edges (indices) adjacent to the face
        c_edges_in_c_faces = np.array(
            [a[b] for a, b in zip(edges_in_c_faces, c_edges_in_faces[c_faces])]
        )

        # Contour edges (indices pointing to the unique edge list)
        c_edge_inds = list(np.flatnonzero(edge_inds))

        # Indices of faces used in contour
        c_face_inds = np.flatnonzero(c_faces)

        # Check gradient of stream function in first (could be any) triangle
        c_face_gradient = g[:, c_face_inds[0]]

        # Vector between two edges in first face used for contour
        vec = (
            points[c_edge_inds.index(c_edges_in_c_faces[0, 0])]
            - points[c_edge_inds.index(c_edges_in_c_faces[0, 1])]
        )

        # Create loop such that it is in the same direction as gradient
        if c_face_gradient.dot(vec) >= 0:
            # Init loop variables
            key = c_edges_in_c_faces[0, 1]
            val = c_edges_in_c_faces[0, 0]
        else:
            # Init loop variables
            key = c_edges_in_c_faces[0, 0]
            val = c_edges_in_c_faces[0, 1]

        ii = 0
        sorted_inds = []
        kmax = len(c_edge_inds)
        k = 0

        # Loop over c_edges by essentially solving a linked list from c_edges_in_c_faces
        while k < kmax:
            sorted_inds.append(c_edge_inds.index(val))
            c_edges_in_c_faces[ii] = -1
            ii, jj = np.nonzero(c_edges_in_c_faces == key)

            if len(ii) == 0:
                # Next edge not found in the adjacency list, contour must be closed now
                # OR the loop is not closed, test this:
                ii_temp, jj_temp = np.nonzero(c_edges_in_c_faces == val)
                if len(ii_temp) > 0:
                    # Proceed to another direction (direction does not matter)
                    ii = ii_temp
                    jj = jj_temp
                else:
                    # Sort points containing contours by adjacency of the edges
                    # and append to contour_polys

                    contour_polys.append(points[sorted_inds])
                    contour_values.append(c)

                    # Break the loop if all edges have been visited
                    if np.all(c_edges_in_c_faces == -1):
                        break

                    # Else find a starting point in another contour at the same level
                    sorted_inds = []

                    ii = np.flatnonzero(c_edges_in_c_faces[:, 0] >= 0)[0]
                    jj = 0

                    # Compute gradient in face
                    c_face_gradient = g[:, c_face_inds[ii]]

                    # once again, check winding direction
                    vec = (
                        points[c_edge_inds.index(c_edges_in_c_faces[ii, 0])]
                        - points[c_edge_inds.index(c_edges_in_c_faces[ii, 1])]
                    )

                    if c_face_gradient.dot(vec) >= 0:
                        jj = 0
                    else:
                        jj = 1

            else:
                # Edge found
                ii = ii[0]
                jj = jj[0]

            # Update key and value
            val = c_edges_in_c_faces[ii, jj]
            key = c_edges_in_c_faces[ii, (jj + 1) % 2]

            k += 1

            if k == kmax:
                raise RuntimeWarning(
                    "Something wrong with the contours, number of max iterations exceeded"
                )
    if return_values:
        return contour_polys, contour_values

    return contour_polys


def simplify_contour(c, min_edge=1e-3, angle_threshold=2e-2, smooth=True):
    """
    Simplifies contours by merging small (short) segments and
    with only a small angle difference.

    Optionally applies smoothing to contour shape.

    Parameters
    ----------
    c: list
        List of polygons describing closed loops.
    min_edge: float
        Minimum edge length. Edges shorter than this are merged.
    angle_threshold: float
        Minimum angle. Edges with smaller angle differences are merged.
    smooth: bool
        If True, apply smoothing to the polygon shapes.

    Returns
    -------
    c: list
        Modified list of polygons

    """
    # Remove small edges by threshold
    vals = [np.ones(c.shape[0]), -np.ones(c.shape[0]), np.ones(c.shape[0])]
    D = spdiags(vals, [1, 0, -c.shape[0] + 1], c.shape[0], c.shape[0])
    edges = D @ c
    c = c[np.linalg.norm(edges, axis=1) > min_edge]
    if len(c) == 0:
        return None

    # Remove nodes on straight lines
    D = spdiags(vals, [1, 0, -c.shape[0] + 1], c.shape[0], c.shape[0])
    H = spdiags(1 / np.linalg.norm(D @ c, axis=1), 0, c.shape[0], c.shape[0])
    DD = H @ D
    c = c[np.linalg.norm(D.T @ DD @ c, axis=-1) > angle_threshold]

    if smooth:
        D = spdiags(vals, [1, 0, -c.shape[0] + 1], c.shape[0], c.shape[0])
        H = spdiags(1 / np.linalg.norm(D @ c, axis=1), 0, c.shape[0], c.shape[0])
        DD = H @ D
        lengths = np.linalg.norm(D @ c, axis=1)
        lengths = 0.5 * abs(D.T) @ lengths  # Mean of edges
        #            c = c - 0.2*lengths[:,None]*(D.T @ DD @ c)
        Nc = c.shape[0]
        c = spsolve(speye(Nc, Nc) + 1.0 * spdiags(lengths, 0, Nc, Nc) @ (D.T @ DD), c)

    return c



"""
Analytic integrals as scalar fields
========================================================

"""

import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh


from bfieldtools.integrals import triangle_potential_uniform
from bfieldtools.integrals import triangle_potential_dipole_linear
from bfieldtools.integrals import gamma0
from bfieldtools.integrals import omega
from bfieldtools.utils import tri_normals_and_areas

#%%
#%% Create there orthogonal planes
points = np.array([[0, 0, 0], [1, 0.5, 0], [0, 1, 0]]) + 0.15

points[:, 2] += 0.1

tris = np.array([[0, 1, 2]])
p_tris = points[tris]
mesh = trimesh.Trimesh(points, tris, process=False)

# Evaluation points
points2 = np.array([[-0.01, 1, 1], [-0.01, 1, 0], [-0.01, 0, 0], [-0.01, 0, 1]]) * 1.2
tris2 = np.array([[0, 1, 2], [2, 3, 0]])
mesh2 = trimesh.Trimesh(points2, tris2)
for ii in range(7):
    mesh2 = mesh2.subdivide()

points3 = np.array([[1, 1, -0.01], [1, 0, -0.01], [0, 0, -0.01], [0, 1, -0.01]]) * 1.2
mesh3 = trimesh.Trimesh(points3, tris2)
for ii in range(7):
    mesh3 = mesh3.subdivide()

points4 = np.array([[1, -0.01, 1], [1, -0.01, 0], [0, -0.01, 0], [0, -0.01, 1]]) * 1.2
mesh4 = trimesh.Trimesh(points4, tris2)
for ii in range(7):
    mesh4 = mesh4.subdivide()


def plot_axes():
    axes_points = np.zeros((3, 3))
    axes = np.eye(3) * 0.3
    vectors = mlab.quiver3d(*axes_points, *axes, mode="arrow", color=(0, 0, 0))
    vectors.glyph.glyph.scale_factor = 1.5
    vectors.glyph.glyph_source.glyph_source.tip_length = 0.1
    vectors.glyph.glyph_source.glyph_source.tip_radius = 0.03
    vectors.glyph.glyph_source.glyph_source.shaft_radius = 0.01


# Difference vectors
RR2 = mesh2.vertices[:, None, None, :] - p_tris[None, :, :, :]
RR3 = mesh3.vertices[:, None, None, :] - p_tris[None, :, :, :]
RR4 = mesh4.vertices[:, None, None, :] - p_tris[None, :, :, :]

tn, ta = tri_normals_and_areas(points, tris)

#%% Plot potentials on the planes and the respective sources on the triangle
for ii, func in enumerate(
    (triangle_potential_uniform, triangle_potential_dipole_linear, gamma0, omega)
):

    mlab.figure(bgcolor=(1, 1, 1))
    # Plot shape and potential
    print(func)
    if ii == 0:
        pot2 = func(RR2, tn)[:, 0]
        pot3 = func(RR3, tn)[:, 0]
        pot4 = func(RR4, tn)[:, 0]
        mlab.triangular_mesh(
            *mesh.vertices.T, mesh.faces, color=(0.5, 0.5, 0.5), opacity=0.7
        )
    if ii == 1:
        pot2 = func(RR2, tn, ta)[:, 0, 1]
        pot3 = func(RR3, tn, ta)[:, 0, 1]
        pot4 = func(RR4, tn, ta)[:, 0, 1]
        mlab.triangular_mesh(
            *mesh.vertices.T, mesh.faces, color=(0.5, 0.5, 0.5), opacity=0.2
        )
        meshq = mesh.copy()
        for jj in range(4):
            meshq = meshq.subdivide()
        u = np.zeros(meshq.vertices.shape)
        r = meshq.vertices
        r2 = np.zeros(r.shape + (3,))
        r2[:, 1] = r
        r2[:, 0] = mesh.vertices[0]
        r2[:, 2] = mesh.vertices[2]
        u[:, 2] = np.linalg.det(r2) / np.linalg.det(mesh.vertices)
        q = mlab.quiver3d(*r.T, *u.T, colormap="gray", mode="arrow")
        q.glyph.glyph.scale_factor = 0.25

    if ii == 2:
        pot2 = func(RR2)[:, 0, 2]
        pot3 = func(RR3)[:, 0, 2]
        pot4 = func(RR4)[:, 0, 2]
        mlab.plot3d(
            *points[0:2].T, color=(0.5, 0.5, 0.5), opacity=0.5, tube_radius=0.02
        )
    if ii == 3:
        pot2 = func(RR2)[:, 0]
        pot3 = func(RR3)[:, 0]
        pot4 = func(RR4)[:, 0]
        mlab.triangular_mesh(
            *mesh.vertices.T, mesh.faces, color=(0.5, 0.5, 0.5), opacity=0.5
        )
        u = np.zeros(meshq.vertices.shape)
        u[:, 2] = 1
        r = meshq.vertices
        q = mlab.quiver3d(*r.T, *u.T, colormap="gray", mode="arrow")
        q.glyph.glyph.scale_factor = 0.2

    plot_axes()

    M = max(max(abs(pot2)), max(abs(pot3)), max(abs(pot4)))
    for m, p in zip((mesh2, mesh3, mesh4), (pot2, pot3, pot4)):
        s = mlab.triangular_mesh(
            *m.vertices.T, m.faces, scalars=p, colormap="seismic", vmin=-M, vmax=M
        )
        s.actor.mapper.interpolate_scalars_before_mapping = True
        s.module_manager.scalar_lut_manager.number_of_colors = 32

    # Move camera a bit
    s.scene.camera.pitch(-2)
    s.scene.camera.azimuth(-25)
    s.scene.camera.elevation(-10)
    s.scene.camera.orthogonalize_view_up()
    s.scene.camera.pitch(1)
    s.scene.camera.orthogonalize_view_up()



import pytest
from bfieldtools import coil_optimize
from .test_mesh_conductor import _fake_mesh_conductor
import cvxpy
import numpy as np

from numpy.testing import assert_allclose


def test_coil_optimize():
    """
    test that optimization works for different mesh_conductor objects and solvers
    """

    for test_mesh in ["unit_sphere", "unit_disc", "plane_w_holes"]:

        for basis_name in ["suh"]:  # , 'inner', 'vertex']:

            c = _fake_mesh_conductor(
                mesh_name=test_mesh, basis_name=basis_name, N_suh=10
            )

            spec = dict(
                coupling=c.B_coupling(np.array([[0, -0.1, 1], [0, 0, 1], [0, 0.1, 1]])),
                target=np.array([[0, 0, 1], [0, 0, 1], [0, 0, 1]]),
                abs_error=0.01,
            )

            for objective in [
                "minimum_inductive_energy",
                "minimum_ohmic_power",
                (0.5, 0.5),
            ]:
                results = []

                results.append(coil_optimize.optimize_lsq(c, [dict(spec)], reg=1e3))

                # For now, test with all solvers that can handle SOC problems
                for solver in [
                    i
                    for i in cvxpy.solvers.defines.INSTALLED_CONIC_SOLVERS
                    if i not in ("GLPK", "GLPK_MI", "SCS")
                ]:
                    results.append(
                        coil_optimize.optimize_streamfunctions(
                            c, [spec], objective, solver
                        )[0]
                    )
                if len(results) > 1:
                    # tolerance is quite high, since some solvers give a bit differing results
                    # in real life, let's not use those solvers.
                    assert_allclose(
                        results[-2],
                        results[-1],
                        rtol=5e-1,
                        atol=0.001 * np.mean(np.abs(results[-2])),
                    )


def test_standalone_functions():
    """
    Tests standalone functions in coil_optimize

    """

    P = 2 * np.array([[2, 0.5], [0.5, 1]])
    q = np.array([1.0, 1.0])
    G = np.array([[-1.0, 0.0], [0.0, -1.0]])
    h = np.array([0.0, 0.0])
    A = np.array([[1.0, 1.0], [1, 2]])
    b = np.array([1.0, 0])

    coil_optimize.cvxopt_solve_qp(P, q)
    coil_optimize.cvxopt_solve_qp(P, q, G, h)
    coil_optimize.cvxpy_solve_qp(P, G, h, solver="CVXOPT")



"""
Legacy code for sphtools
"""


def compute_sphcoeffs_mesh(mesh, lmax):
    """
    Computes multipole moment (spherical harmonics coefficient) transformation
    from the mesh.

    Parameters
    ----------
    mesh: mesh object - the surface mesh
    lmax: int
        maximum degree l of the fit

    Returns
    -------
    alm: (lmax*(lmax+2)xNvertices array
          transformation from the mesh to alm coefficients (r**l-terms)
    blm: (lmax*(lmax+2)xNvertices array
          transformation from the mesh to blm coefficients (r**(-l)-terms)

    """

    Gx, Gy, Gz = gradient_matrix(mesh, rotated=True)
    tri_normals, tri_areas = tri_normals_and_areas(mesh.vertices, mesh.faces)

    centers = np.mean(mesh.vertices[mesh.faces], axis=1)
    centers_sp = cartesian2spherical(centers)

    alm = np.zeros((lmax * (lmax + 2), mesh.vertices.shape[0]))
    blm = np.zeros((lmax * (lmax + 2), mesh.vertices.shape[0]))

    idx = 0
    for l in range(1, lmax + 1):
        for m in range(-1 * l, l + 1):
            derylm = np.sqrt(l * (l + 1)) * Blm(
                l, m, centers_sp[:, 1], centers_sp[:, 2]
            )
            derylm = sphvec2cart(centers_sp, derylm)
            # FIX: gradient of Ylm has also 1/r multiplier
            derylm /= centers_sp[:, 0:1]

            crossp = np.cross(centers, derylm)
            alm_terms = crossp.T * (centers_sp[:, 0] ** l) * tri_areas
            blm_terms = crossp.T * (centers_sp[:, 0] ** (-l - 1)) * tri_areas
            #            integral_alm = np.sum(G*alm_terms[:,:,None], axis=(0,1))
            #            integral_blm = np.sum(G*blm_terms[:,:,None], axis=(0,1))

            integral_alm = alm_terms[0] @ Gx + alm_terms[1] @ Gy + alm_terms[2] @ Gz
            integral_blm = blm_terms[0] @ Gx + blm_terms[1] @ Gy + blm_terms[2] @ Gz

            # FIX: l-coefficients here were swapped
            blm[idx] = 1 / ((2 * l + 1) * l) * integral_blm
            # FIX: REMOVED MINUS SIGN
            alm[idx] = (
                -1 / ((2 * l + 1) * (l + 1)) * integral_alm
            )  # THIS SIGN SHOULD BE CHECKED TOO
            #            for i in range(mesh.vertices.shape[0]):
            #                G = np.zeros(crossp.shape)
            #                G[:,0] = Gx[:,i]
            #                G[:,1] = Gy[:,i]
            #                G[:,2] = Gz[:,i]
            #                dotp = np.sum(G*crossp,axis=1)
            #                integral_blm = np.sum(dotp*centers_sp[:,0]**l*tri_areas)
            #                blm[idx,i] = -1/((2*l+1)*(l+1))*integral_blm
            #
            #                integral_alm = np.sum(dotp*centers_sp[:,0]**(-l-1)*tri_areas)
            #                alm[idx,i] = 1/((2*l+1)*l)*integral_alm

            idx += 1
        print("l = %d computed" % (l))

    return alm, blm



"""
High-order spherical harmonic biplanar coil design
==================================================

Example showing a basic biplanar coil producing a high-order spherical harmonic field
in a specific target region between the two coil planes.

"""

import numpy as np
from mayavi import mlab
import trimesh


from bfieldtools.mesh_conductor import MeshConductor
from bfieldtools.coil_optimize import optimize_streamfunctions
from bfieldtools.utils import combine_meshes, load_example_mesh


# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")

# Specify coil plane geometry
center_offset = np.array([0, 0, 0])
standoff = np.array([0, 3, 0])

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

joined_planes = combine_meshes((coil_plus, coil_minus))

# Create mesh class object
coil = MeshConductor(
    mesh_obj=joined_planes, fix_normals=True, basis_name="suh", N_suh=100
)

#%%
# Set up target and stray field points

# Here, the target points are on a volumetric grid within a sphere

center = np.array([0, 0, 0])

sidelength = 1.5
n = 8
xx = np.linspace(-sidelength / 2, sidelength / 2, n)
yy = np.linspace(-sidelength / 2, sidelength / 2, n)
zz = np.linspace(-sidelength / 2, sidelength / 2, n)
X, Y, Z = np.meshgrid(xx, yy, zz, indexing="ij")

x = X.ravel()
y = Y.ravel()
z = Z.ravel()

target_points = np.array([x, y, z]).T

# Turn cube into sphere by rejecting points "in the corners"
target_points = (
    target_points[np.linalg.norm(target_points, axis=1) < sidelength / 2] + center
)


#%%
# Create bfield specifications used when optimizing the coil geometry


from bfieldtools import sphtools


lmax = 4
alm = np.zeros((lmax * (lmax + 2),))
blm = np.zeros((lmax * (lmax + 2),))

# Set one specific component to one
blm[16] += 1

sphfield = sphtools.field(target_points, alm, blm, lmax)

target_field = sphfield / np.max(sphfield[:, 0])


coil.plot_mesh(opacity=0.2)
mlab.quiver3d(*target_points.T, *sphfield.T)


target_spec = {
    "coupling": coil.B_coupling(target_points),
    "abs_error": 0.1,
    "target": target_field,
}


#%%
# Run QP solver
import mosek

coil.s, prob = optimize_streamfunctions(
    coil,
    [target_spec],
    objective="minimum_inductive_energy",
    solver="MOSEK",
    solver_opts={"mosek_params": {mosek.iparam.num_threads: 8}},
)

#%%
# Plot coil windings and target points
coil.s.discretize(N_contours=10).plot_loops()



"""
Integrals testing
==================================================

"""

import numpy as np

import matplotlib.pyplot as plt

import trimesh
from mayavi import mlab

#%%
# Test potential shape slightly above the surface
#%%
x = np.sin(np.pi / 6)
y = np.cos(np.pi / 6)
points = (
    np.array(
        [
            [0, 0, 0],
            [1, 0, 0],
            [x, y, 0],
            [-x, y, 0],
            [-1, 0, 0],
            [-x, -y, 0],
            [x, -y, 0],
        ]
    )
    * 2
)

tris = np.array([[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 5], [0, 5, 6], [0, 6, 1]])
mesh = trimesh.Trimesh(points, tris)
scalars = np.zeros(7)
scalars[0] = 1

#%%
# Linear dipole density
#%%

# Sign ok
points = np.array([[0.1, 1, 1], [0.1, 1, -1], [0.1, -1, -1], [0.1, -1, 1]]) * 2
# points = np.roll(points, 2, 1)
tris = np.array([[0, 1, 2], [2, 3, 0]])
mesh2 = trimesh.Trimesh(points, tris)
for ii in range(7):
    mesh2 = mesh2.subdivide()

from bfieldtools.legacy.integrals import triangle_potential_dipole_linear as t1
from bfieldtools.integrals import triangle_potential_dipole_linear as t2

RR = mesh2.vertices[:, None, None, :] - mesh.vertices[None, mesh.faces]
p1 = t1(RR, mesh.face_normals, mesh.area_faces, planar=False)
p2 = t2(RR, mesh.face_normals, mesh.area_faces)

assert np.allclose(p1, p2)


mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=p1[:, :, 0].sum(axis=1))
mlab.colorbar()
mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, representation="wireframe")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)

mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=p2[:, :, 0].sum(axis=1))
mlab.colorbar()
mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, representation="wireframe")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)

mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
mlab.triangular_mesh(
    *mesh2.vertices.T, mesh2.faces, scalars=(p1 - p2)[:, :, 0].sum(axis=1)
)
mlab.colorbar()

#%%
#
points = np.zeros((100, 3))
points[:, 2] = np.linspace(-1, 1, 100)
from bfieldtools.legacy.integrals import omega as omega1
from bfieldtools.integrals import omega as omega2

RR = points[:, None, None, :] - mesh.vertices[None, mesh.faces]
o1 = omega1(RR).sum(axis=1)
o2 = omega2(RR).sum(axis=1)

assert np.allclose(o1, -o2)

plt.plot(o1)
plt.plot(o2)
mlab.plot3d(*points.T, points[:, 2], colormap="seismic")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)

#%%
# Plot x_i

from bfieldtools.integrals import x_distance

RR = mesh2.vertices[:, None, None, :] - mesh.vertices[None, mesh.faces]
xdist = x_distance(RR, mesh.face_normals)
mlab.triangular_mesh(
    *mesh2.vertices.T,
    mesh2.faces,
    scalars=xdist[:, 1, 0],
    vmin=-1,
    vmax=1,
    colormap="seismic"
)
mlab.colorbar()
mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, representation="wireframe")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)

#%%
# Uniform charge density
from bfieldtools.legacy.integrals import triangle_potential_uniform as u1
from bfieldtools.integrals import triangle_potential_uniform as u2

RR = mesh2.vertices[:, None, None, :] - mesh.vertices[None, mesh.faces]
p1 = u1(RR, mesh.face_normals, planar=False)
p2 = u2(RR, mesh.face_normals, planar=False)

assert np.allclose(p1, p2)


mlab.figure("uniform charge density (old)")
mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=p1.sum(axis=1))
mlab.colorbar()
mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, representation="wireframe")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)
mlab.figure("uniform charge density (new)")
mlab.triangular_mesh(*mesh2.vertices.T, mesh2.faces, scalars=p2.sum(axis=1))
mlab.colorbar()
mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, representation="wireframe")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)

#%%
#
from bfieldtools.integrals import d_distance

RR = mesh2.vertices[:, None, None, :] - mesh.vertices[None, mesh.faces]
ddist = d_distance(RR, mesh.face_normals)
mlab.figure("d distance")
mlab.triangular_mesh(
    *mesh2.vertices.T,
    mesh2.faces,
    scalars=ddist[:, 0],
    vmin=-1,
    vmax=1,
    colormap="seismic"
)
mlab.colorbar()
mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, representation="wireframe")
mlab.quiver3d(*mesh.triangles_center.T, *mesh.face_normals.T)

#%%
#
from bfieldtools.legacy.mesh_magnetics import (
    magnetic_field_coupling_analytic as magnetic_field_coupling_analytic_old,
)
from bfieldtools.mesh_magnetics import magnetic_field_coupling_analytic

b1 = magnetic_field_coupling_analytic_old(mesh, mesh2.vertices)
b2 = magnetic_field_coupling_analytic(mesh, mesh2.vertices)

assert np.allclose(b1, b2)

mlab.figure("b field")
mlab.quiver3d(*mesh2.vertices.T, *b1[:, :, 0].T)
mlab.quiver3d(*mesh2.vertices.T, *b2[:, :, 0].T)

#%%
# Gammma
from bfieldtools.legacy.integrals import gamma0 as g1
from bfieldtools.integrals import gamma0 as g2

# RR =  mesh2.vertices[:, None, None, :] - mesh.vertices[None, mesh.faces]
t = np.linspace(-1.5, 1.5)
points = (
    t[:, None] * mesh.vertices[mesh.faces][0][0]
    + (1 - t)[:, None] * mesh.vertices[mesh.faces][0][1]
)


R = points[:, None, None, :] - mesh.vertices[None, mesh.faces]
p1 = g1(R, symmetrize=True)
p2 = g2(R, symmetrize=True)

assert np.allclose(p1, p2)

plt.figure()
plt.plot(p1[:, 0, :])
plt.figure()
plt.plot(p2[:, 0, :])



import numpy as np
from mayavi import mlab

from bfieldtools import suhtools
from bfieldtools.utils import load_example_mesh

from .test_mesh_conductor import _fake_mesh_conductor

import pytest

from numpy.testing import assert_allclose


def test_suhbasis():

    mesh = load_example_mesh("unit_disc")

    for obj in [mesh, _fake_mesh_conductor(mesh_name="unit_disc")]:
        for mag in [False]:
            for bc in ["neumann", "dirichlet"]:
                suh = suhtools.SuhBasis(obj, Nc=10, boundary_condition=bc, magnetic=mag)
                assert suh.basis.shape[1] == 10

                suh.field(suh.basis[1], np.array([[0, 0, 1]]))

                suh.plot(Nfuncs=3)

    suh = suhtools.SuhBasis(mesh, Nc=10, boundary_condition="dirichlet", magnetic="DC")
    assert suh.basis.shape[1] == 10

    suh = suhtools.SuhBasis(mesh, Nc=10, boundary_condition="dirichlet", magnetic="AC")
    assert suh.basis.shape[1] == 10

    suh = suhtools.SuhBasis(mesh, boundary_condition="dirichlet", magnetic="DC")
    suh = suhtools.SuhBasis(mesh, boundary_condition="dirichlet", magnetic="AC")
    suh = suhtools.SuhBasis(mesh, boundary_condition="dirichlet", magnetic=False)

    suh.calculate_basis(shiftinvert=False, v0=np.ones(57))

    try:
        suh = suhtools.SuhBasis(10, Nc=10)
        assert suh.basis.shape[1] == 10
    except:
        print("Caught test exception")

    f = mlab.figure()

    suh.plot([0, 2], figure=f)
    suh.plot(Nfuncs=4, Ncols=2, figure=f)


def test_shiftinvert():

    mesh = load_example_mesh("unit_disc")

    suh = suhtools.SuhBasis(mesh, Nc=10, boundary_condition="dirichlet", magnetic=False)

    suh.calculate_basis(shiftinvert=True)
    shiftinvert_eigenvals = suh.eigenvals
    shiftinvert_basis = suh.basis

    suh.calculate_basis(shiftinvert=False)
    nonshiftinvert_eigenvals = suh.eigenvals
    nonshiftinvert_basis = suh.basis

    assert_allclose(shiftinvert_eigenvals, nonshiftinvert_eigenvals)


def test_suhbasis_closed():
    import trimesh

    # Low-res ico-sphere, but need to copy since Trimesh primitives are protected
    mesh = trimesh.primitives.Sphere(subdivisions=2)
    mesh = trimesh.Trimesh(mesh.vertices, mesh.faces)

    for mag in [False, "DC", "AC"]:
        suh = suhtools.SuhBasis(mesh, Nc=10, magnetic=mag)
        assert suh.basis.shape[1] == 10

        suh = suhtools.SuhBasis(mesh, magnetic=mag)
        assert suh.basis.shape[1] == suh.Nc

    suh = suhtools.SuhBasis(mesh, Nc=41, magnetic=mag)
    assert suh.basis.shape[1] == suh.Nc


@pytest.mark.xfail
def test_suhbasis_fail():
    mesh = load_example_mesh("unit_disc")
    suh = suhtools.SuhBasis(mesh, Nc=10, boundary_condition="foo")


@pytest.mark.xfail
def test_suhbasis_fail2():
    mesh = load_example_mesh("unit_disc")
    suh = suhtools.SuhBasis(mesh, Nc=10, magnetic="foo")


@pytest.mark.xfail
def test_suhbasis_fail3():
    mesh = load_example_mesh("unit_disc")
    suh = suhtools.SuhBasis(mesh, Nc=1000)



"""
Includes files for coil optimization (stream function optimization)
using either a numerical solver or regularized least squares

"""

__all__ = [
    "cvxopt_solve_qp",
    "cvxpy_solve_qp",
    "optimize_lsq",
    "optimize_streamfunctions",
]

import numpy as np
import cvxopt
from cvxopt import matrix
from scipy.sparse.linalg import svds
from scipy.linalg import eigh
from scipy.linalg import eigvalsh
import cvxpy as cp

from .mesh_conductor import StreamFunction


def cvxpy_solve_qp(P, G, h, solver=cp.MOSEK, tolerance=None):
    """
    Bare-bones quadratic programming solver function for CVXPY.
    
    Minimizes

    .. math::
        (1/2) \mathbf{x}^T  \mathbf{P}  \mathbf{x}

    subject to

    .. math::
        \mathbf{G} \mathbf{x} \leq \mathbf{h}


    Parameters
    ----------

    P: (N_x, N_x) array
        Quadratic minimization matrix
    G: (N_h, N_x) array
        Linear inequality matrix
    h: (N_h, ) array
        Linear inequality constraint vector
    solver: string or cvxpy solver
        solver to use in CVXPY, can be passed as string or directly
    tolerance: float or None (default)
        Override default tolerance values

    Returns
    -------
    x.value: (N_x, ) array
        Optimized solution for x (if available)
    prob: CVXPY problem
        Problem object with optimization info

    """

    P = 0.5 * (P + P.T)

    x = cp.Variable(len(P))

    objective = cp.Minimize((1 / 2) * cp.quad_form(x, P))

    constraints = [G @ x <= h]

    prob = cp.Problem(objective, constraints)
    if tolerance is None:
        prob.solve(solver=solver, verbose=True)  # Returns the optimal value.
    elif solver == cp.CVXOPT:
        prob.solve(
            solver=solver,
            verbose=True,
            abstol=tolerance,
            feastol=tolerance,
            reltol=tolerance,
        )  # Returns the optimal value.
    elif solver == cp.SCS:
        prob.solve(
            solver=solver, verbose=True, eps=tolerance
        )  # Returns the optimal value.

    # Print result.
    print("\nThe optimal value is", prob.value)
    print("A solution x is")
    print(x.value)
    print("A dual solution corresponding to the inequality constraints is")
    print(prob.constraints[0].dual_value)

    return x.value, prob


def cvxopt_solve_qp(P, q, G=None, h=None, A=None, b=None, tolerance=1e-7, **kwargs):
    """
    Use cvxopt (without CVXPY wrapper) for quadratic programming.
    
    Minimize

    .. math::
        (1/2) \mathbf{x}^T  \mathbf{P}  \mathbf{x} + \mathbf{q}^T \mathbf{x}

    subject to

    .. math::
        \mathbf{G} \mathbf{x} \leq \mathbf{h}

    and

    .. math::
        \mathbf{A} \mathbf{x} = \mathbf{b}

    Parameters
    ----------

    P: (N_x, N_x) array
        Quadratic minimization matrix
    q: (N_x, ) array
        Linear penalty term vector
    G: (N_h, N_x) array
        Linear inequality matrix
    h: (N_h, ) array
        Linear inequality constraint vector
    A: (N_b, N_x) array
        Linear equality matrix
    b: (N_b, ) array
        Linear equality constraint vector
    **kwargs
        Use to pass cvxopt solver options, such as tolerance

    Returns
    -------
    x: (N_x, ) array
        Optimized solution for x (if available)
    sol: CVXOPT solution
        Solution object with optimization info

    """

    P = 0.5 * (P + P.T)  # make sure P is symmetric

    args = [matrix(P), matrix(q)]
    if G is not None:
        args.extend([matrix(G), matrix(h)])
        if A is not None:
            args.extend([matrix(A), matrix(b)])

    for key, val in kwargs:
        cvxopt.solver.options[key] = val

    sol = cvxopt.solvers.qp(*args)
    if "optimal" not in sol["status"]:
        return None, sol
    return np.array(sol["x"]).reshape((P.shape[1],)), sol


def optimize_streamfunctions(
    mesh_conductor,
    bfield_specification,
    objective="minimum_inductive_energy",
    solver=None,
    solver_opts={},
    problem=None,
):
    """
    Quadratic optimization of coil stream function according to a specified objective.

    Utilizes CVXPY and a numerical iterative solver.

    Parameters
    ----------
    mesh_conductor: MeshConductor object
        Contains Trimesh mesh as well as physical properties, e.g. inductance
    bfield_specification: list
        List in which element is a dictionary containing a coil specification.
        See notes for specification syntax.
    objective: string or dict
        if string, either *'minimum_inductive_energy'* or *'minimum_ohmic_power'*
        if tuple, should contain: (a, b), where a and b are floats describing the
        inductive and resitive weighting factors.
        The resistance matrix is scaled according to the largest singular value
        of the inductance matrix for consistent behavior across meshes.
    solver: string
        string specifying which solver CVXPY will use
    solver_opt: dict
        dict containing solver options CVXPY will pass to the solver
    problem: CVXPY problem object
        If passed, will use already existing problem (**MUST BE SAME DIMENSIONS**) to
        skip DCP processing/reformulation time.

    Returns
    -------
    s: vector
        Vector with length len(`mesh_conductor.mesh.vertices`), containing the
        optimized current density values at each mesh vertex
    prob: CVXPY problem object
        CVXPY problem object containing data, formulation, solution, metric etc

    Notes
    -----

    Each specification is a dict, which contains
     - coupling: Coupling matrix (N_r, N_verts, 3)
     - target: (N_r, 3)
     - abs_error: float or (N_r, 3)
     - rel_error: float or (N_r, 3)
    Either abs_error, rel_error or both must be present.

    """

    if objective == "minimum_inductive_energy":
        objective = (1, 0)
    elif objective == "minimum_ohmic_power":
        objective = (0, 1)

    quadratic_matrix = _construct_quadratic_objective(objective, mesh_conductor)

    constraint_matrix, upper_bounds, lower_bounds = _construct_constraints(
        mesh_conductor, bfield_specification
    )
    # Compute, scale constraint matrix according to largest singular value
    u, s, vt = svds(constraint_matrix, k=1)

    # If no pre-constructed problem is passed, create it
    if problem is None:
        print("Pre-existing problem not passed, creating...")
        # Symbolic variable for CVXPY
        x = cp.Variable(shape=(len(quadratic_matrix),), name="x")

        # Parameters into which data is loaded
        # P does not need to be PSD when in this formulation
        # It should be full rank, however
        P = cp.Parameter(shape=quadratic_matrix.shape, name="P")  # , PSD=True)
        G = cp.Parameter(shape=constraint_matrix.shape, name="G")
        lb = cp.Parameter(shape=lower_bounds.shape, name="lb")
        ub = cp.Parameter(shape=upper_bounds.shape, name="ub")

        # Formulate problem and constraints
        objective = cp.Minimize((1 / 2) * cp.sum_squares(P @ x))

        constraints = [G @ x >= lb, G @ x <= ub]

        problem = cp.Problem(objective, constraints)
    else:
        print("Existing problem passed")

    # Assign values to parameters
    print("Passing parameters to problem...")
    for par in problem.parameters():
        if par.name() == "P":
            # Make sure that quadratic matrix is positive semi-definite, scale constraint matrix
            # par.value = sqrtm(0.5 * (quadratic_matrix + quadratic_matrix.T))
            PP = 0.5 * (quadratic_matrix + quadratic_matrix.T)
            par.value = np.linalg.cholesky(PP).T
        elif par.name() == "G":
            par.value = constraint_matrix / s[0]
        elif par.name() == "lb":
            par.value = lower_bounds
        elif par.name() == "ub":
            par.value = upper_bounds
        else:
            print("Unknown parameter encountered")

    # Run solver
    print("Passing problem to solver...")
    problem.solve(solver=solver, verbose=True, **solver_opts)

    # extract optimized streamfunction, scale by same singular value as constraint matrix
    S = StreamFunction(
        problem.variables()[0].value / s[0], mesh_conductor=mesh_conductor
    )

    return S, problem


def optimize_lsq(
    mesh_conductor, bfield_specification, reg=1e3, objective="minimum_inductive_energy"
):
    """
    Optimization of coil stream function according to a specified objective using least-squares.
    
    Parameters
    ----------
    mesh_conductor: MeshConductor object
        Contains Trimesh mesh as well as physical properties, e.g. inductance
    bfield_specification: list
        List in which element is a dictionary containing a coil specification.
    objective: string or dict
        if string, either *'minimum_inductive_energy'* or *'minimum_ohmic_power'*
        if tuple, should contain: (a, b), where a and b are floats describing the
        inductive and resitive weighting factors.
        The resistance matrix is scaled according to the largest singular value
        of the inductance matrix for consistent behavior across meshes.
    reg: float
        Regularization/tradeoff parameter (lambda). A larger lambda leads to
        more emphasis on the specification, at the cost of the quadratic objective.
        The lambda value is relative to the maximum singular value w
        of the eigenvalue equation

        .. math::
            \mathbf{C}^T \mathbf{C}  \mathbf{v}[:,i] = \mathbf{w}[i]  \mathbf{Q} \mathbf{v}[:,i]

        where C is the constraint matrix and Q is the quadratic objective matrix.

    Returns
    -------
    S: StreamFunction
        Optimization solution

    Notes
    -----

    Each specification is a dict, which contains
     - coupling: Coupling matrix (N_r, N_verts, 3)
     - target: (N_r, 3)

    **NOTE** The following spec parameters are ignored:
     - abs_error: float or (N_r, 3)
     - rel_error: float or (N_r, 3)
    """

    if objective == "minimum_inductive_energy":
        objective = (1, 0)
    elif objective == "minimum_ohmic_power":
        objective = (0, 1)

    quadratic_matrix = _construct_quadratic_objective(objective, mesh_conductor)

    # Make sure that quadratic matrix is positive semi-definite
    quadratic_matrix = 0.5 * (quadratic_matrix + quadratic_matrix.T)

    print("Error tolerances in specification will be ignored when using lsq")
    for spec in bfield_specification:
        spec.pop("abs_error", None)
        spec.pop("rel_error", None)

    constraint_matrix, target = _construct_constraints(
        mesh_conductor, bfield_specification
    )

    # Compute, scale constraint matrix according to largest singular value
    u, s, vt = svds(constraint_matrix, k=1)
    constraint_matrix /= s[0]

    ss_max = eigvalsh(
        constraint_matrix.T @ constraint_matrix,
        quadratic_matrix,
        eigvals=[quadratic_matrix.shape[1] - 1, quadratic_matrix.shape[1] - 1],
    )

    S = np.linalg.solve(
        constraint_matrix.T @ constraint_matrix + ss_max / reg * quadratic_matrix,
        constraint_matrix.T @ target,
    )

    return StreamFunction(S / s[0], mesh_conductor)


def _construct_constraints(mesh_conductor, bfield_specification):
    """
    Stacks together constraint and coupling matrices for stream function optimization

    """
    # Initialize inequality constraint matrix and constraints
    constraint_matrix = np.zeros((0, mesh_conductor.basis.shape[1]))
    upper_bound_stack = np.zeros((0,))
    lower_bound_stack = np.zeros((0,))

    target_stack = np.zeros((0,))

    # Populate inequality constraints with bfield specifications
    for spec in bfield_specification:

        # Reshape so that values on axis 1 are x1, y1, z1, x2, y2, z2, etc.
        # If not 3D matrix, assuming the use of spherical harmonics
        if spec["coupling"].ndim == 3:
            C = spec["coupling"].transpose((2, 0, 1))
            C = C.reshape((C.shape[0], -1)).T
        else:
            C = spec["coupling"]

        # Append specification to constraint matrix and bounds
        constraint_matrix = np.append(constraint_matrix, C, axis=0)

        if "rel_error" in spec or "abs_error" in spec:
            # Apply relative error to bounds
            if "rel_error" in spec:
                upper_bound = spec["target"] * (
                    1 + np.sign(spec["target"]) * spec["rel_error"]
                )
                lower_bound = spec["target"] * (
                    1 - np.sign(spec["target"]) * spec["rel_error"]
                )

                # If present, also apply absolute error
                if "abs_error" in spec:
                    upper_bound += spec["abs_error"]
                    lower_bound -= spec["abs_error"]

            # Apply absolute error to bounds
            else:
                upper_bound = spec["target"] + spec["abs_error"]
                lower_bound = spec["target"] - spec["abs_error"]

            # Flatten to match C matrix
            upper_bound = upper_bound.flatten()
            lower_bound = lower_bound.flatten()

            # Append to stack
            upper_bound_stack = np.append(upper_bound_stack, upper_bound, axis=0)
            lower_bound_stack = np.append(lower_bound_stack, lower_bound, axis=0)

        else:
            target = spec["target"]

            # Flatten to match C matrix
            target = target.flatten()

            # Append to stack
            target_stack = np.append(target_stack, target, axis=0)

    if "rel_error" in spec or "abs_error" in spec:
        return constraint_matrix, upper_bound_stack, lower_bound_stack

    return constraint_matrix, target_stack


def _construct_quadratic_objective(objective, mesh_conductor, deflate=True):
    """

    """
    # Construct quadratic objective matrix
    if objective == (1, 0):

        quadratic_matrix = mesh_conductor.inductance

    elif objective == (0, 1):

        quadratic_matrix = mesh_conductor.resistance

    elif isinstance(objective, tuple):

        L = mesh_conductor.inductance

        R = mesh_conductor.resistance

        print(
            "Scaling inductance and resistance matrices before optimization.\
              This requires eigenvalue computation, hold on."
        )

        max_eval_L = eigh(L, eigvals=(L.shape[0] - 1, L.shape[0] - 1))[0][0]
        max_eval_R = eigh(R, eigvals=(L.shape[0] - 1, L.shape[0] - 1))[0][0]

        scaled_R = max_eval_L / max_eval_R * R

        quadratic_matrix = objective[0] * L + objective[1] * scaled_R
    else:
        print("Custom objective passed, assuming it is a matrix of correct dimensions")
        quadratic_matrix = objective

    # Scale whole quadratic term according to largest eigenvalue
    max_eval_quad = eigh(
        quadratic_matrix,
        eigvals=(quadratic_matrix.shape[0] - 1, quadratic_matrix.shape[0] - 1),
    )[0][0]

    quadratic_matrix /= max_eval_quad

    if deflate:
        evals = eigh(quadratic_matrix, eigvals=(0, 1), eigvals_only=True)
        if abs(evals[0]) < 1e-8 * abs(evals[1]):
            quadratic_matrix += np.ones / np.sqrt(quadratic_matrix.shape[0])

    return quadratic_matrix



"""

Compare SUH and SPH basis functions for the magnetic field
=============================================================

"""


import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh
import pkg_resources
from pyface.api import GUI

_gui = GUI()


from bfieldtools.mesh_magnetics import magnetic_field_coupling
from bfieldtools.mesh_magnetics import magnetic_field_coupling_analytic
from bfieldtools.mesh_magnetics import scalar_potential_coupling
from bfieldtools.sphtools import compute_sphcoeffs_mesh, basis_fields
from bfieldtools.suhtools import SuhBasis
from bfieldtools.utils import load_example_mesh


mesh = load_example_mesh("bunny_repaired")

mesh.vertices -= mesh.vertices.mean(axis=0)

mesh_field = mesh.copy()
mesh_field.vertices += 0.005 * mesh_field.vertex_normals
mesh_field = trimesh.smoothing.filter_laplacian(mesh_field, iterations=1)

Ca, Cb = basis_fields(mesh_field.vertices, 4)

bsuh = SuhBasis(mesh, 25)
Csuh = magnetic_field_coupling_analytic(mesh, mesh_field.vertices) @ bsuh.basis

#%%
def plot_basis_fields(C, comps):
    d = 0.17
    i = 0
    j = 0
    for n in comps:
        p = 1.05 * mesh_field.vertices.copy()
        p2 = mesh_field.vertices.copy()
        #        p[:,1] -= i*d
        #        p2[:,1] -= i*d
        p[:, 0] += i * d
        p2[:, 0] += i * d
        m = np.max(np.linalg.norm(C[:, :, n], axis=0))
        vectors = mlab.quiver3d(
            *p.T, *C[:, :, n].T, mode="arrow", colormap="Greys", vmin=0, vmax=m
        )
        vectors.glyph.mask_input_points = True
        vectors.glyph.mask_points.maximum_number_of_points = 1800
        vectors.glyph.mask_points.random_mode_type = 1
        vectors.glyph.glyph_source.glyph_position = "center"
        vectors.glyph.glyph_source.glyph_source.shaft_radius = 0.02
        vectors.glyph.glyph_source.glyph_source.tip_radius = 0.06
        vectors.glyph.glyph.scale_factor = 0.03
        #        m = np.max(abs((C[:,:,n].T*mesh_field.vertex_normals.T).sum(axis=0)))
        #        s =mlab.triangular_mesh(*p.T, mesh_field.faces,
        #                             scalars=(C[:,:,n].T*mesh_field.vertex_normals.T).sum(axis=0),
        #                             colormap='seismic', vmin=-m, vmax=m, opacity=0.7)
        #        s.actor.property.backface_culling = True
        m = np.max(abs((C[:, :, n].T * mesh_field.vertex_normals.T).sum(axis=0)))
        s = mlab.triangular_mesh(
            *p2.T,
            mesh.faces,
            scalars=(C[:, :, n].T * mesh_field.vertex_normals.T).sum(axis=0),
            colormap="bwr",
            vmin=-m,
            vmax=m
        )
        s.actor.mapper.interpolate_scalars_before_mapping = True
        s.module_manager.scalar_lut_manager.number_of_colors = 15
        i += 1


comps = [0, 4, 10, 15]
scene = mlab.figure(bgcolor=(1, 1, 1), size=(1200, 350))
plot_basis_fields(Ca, comps)
scene.scene.parallel_projection = True
scene.scene.z_plus_view()
scene.scene.camera.zoom(4)
while scene.scene.light_manager is None:
    _gui.process_events()
scene.scene.light_manager.lights[2].intensity = 0.2


scene = mlab.figure(bgcolor=(1, 1, 1), size=(1200, 350))
plot_basis_fields(Csuh, comps)
scene.scene.parallel_projection = True
scene.scene.z_plus_view()
scene.scene.camera.zoom(4)
while scene.scene.light_manager is None:
    _gui.process_events()
scene.scene.light_manager.lights[2].intensity = 0.2


#%% test representation
# from bfieldtools.mesh_magnetics import scalar_potential_coupling
# scaling_factor = 0.02
##Load simple plane mesh that is centered on the origin
# file_obj=pkg_resources.resource_filename('bfieldtools', 'example_meshes/10x10_plane_hires.obj')
# plane = trimesh.load(file_obj=file_obj, process=False)
# plane.apply_scale(scaling_factor)
## Rotate to x-plane
# t = np.eye(4)
# t[1:3,1:3] = np.array([[0,1],[-1,0]])
# plane.apply_transform(t)
# plane = plane.subdivide()
#
# u = scalar_potential_coupling(mesh, plane.vertices, multiply_coeff=False)
# mask = 1 + np.sum(u, axis=1)/(4*np.pi)
# mask[mask < 1e-6]  = 0
#
#
# Ca, Cb = basis_fields(plane.vertices, 6)
# bsuh = SuhBasis(mesh, 48)
# CB = magnetic_field_coupling_analytic(mesh, plane.vertices)
# Csuh = CB @ bsuh.basis
#
# A,B = compute_sphcoeffs_mesh(mesh, 6)
# b1 = np.einsum('ij,lik->lkj', A , Ca) # Mapping from verts to alpha to field
# b2 = np.einsum('ij,klj->lki', bsuh.mass @ bsuh.basis  , Csuh) # Mapping from verts to suh to field
#
# ind=1932
# scalars = np.zeros(mesh.vertices.shape[0])
# scalars[ind] = 1
# mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, scalars=scalars, colormap='Blues')
# surf= mlab.triangular_mesh(*plane.vertices.T, plane.faces, colormap='viridis',
#                     scalars = mask*np.linalg.norm(CB[:,:,ind], axis=1))
# surf.actor.mapper.interpolate_scalars_before_mapping = True
# surf.module_manager.scalar_lut_manager.number_of_colors = 16
#
# mlab.figure()
# mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, scalars=scalars, colormap='Blues')
# surf = mlab.triangular_mesh(*plane.vertices.T, plane.faces, colormap='viridis',
#                     scalars = mask*np.linalg.norm(b1[:,:,ind], axis=0))
# surf.actor.mapper.interpolate_scalars_before_mapping = True
# surf.module_manager.scalar_lut_manager.number_of_colors = 16
#
# mlab.figure()
# mlab.triangular_mesh(*mesh.vertices.T, mesh.faces, scalars=scalars, colormap='Blues')
# surf = mlab.triangular_mesh(*plane.vertices.T, plane.faces, colormap='viridis',
#                     scalars = mask*np.linalg.norm(b2[:,:,ind], axis=0))
# surf.actor.mapper.interpolate_scalars_before_mapping = True
# surf.module_manager.scalar_lut_manager.number_of_colors = 16



from bfieldtools import integrals
from bfieldtools.mesh_calculus import mass_matrix
from bfieldtools.utils import load_example_mesh

import pytest

import numpy as np
from numpy.testing import assert_allclose
from scipy.sparse import csc_matrix


def test_gamma0():
    """
    No specific test done for this
    """

    pass


def test_omega():
    """
    Tests solid angle (whether sphere sums up to 4*pi)

    """

    mesh = load_example_mesh("unit_sphere")

    points = np.array([[0, 0, 0], [0.1, 0, 0.2], [-0.1, 0.1, -0.2]])

    R = points[:, None, None, :] - mesh.vertices[mesh.faces][None, :, :, :]

    solid_angle = integrals.omega(R)

    assert_allclose(np.sum(solid_angle, axis=1), 4 * np.pi)


def test_x_distance():

    import trimesh

    vertices = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [-1, 0, 0]])

    faces = np.array([[2, 1, 0], [0, 3, 2]])

    mesh = trimesh.Trimesh(vertices=vertices, faces=faces)

    points = np.array([[0, 1, 0]])

    R = points[:, None, None, :] - mesh.vertices[faces][None, :, :, :]

    x = integrals.x_distance(R, mesh.face_normals, mesh.area_faces)

    assert_allclose(x, np.array([[[0.0, 1.0, 0.0], [2.0, -1.0, -1.0]]]))


def test_d_distance():

    mesh = load_example_mesh("unit_disc")

    for z_offset in [1, 0, -1]:
        points = mesh.vertices + np.array([[0, 0, z_offset]])

        R = points[:, None, None, :] - mesh.vertices[mesh.faces][None, :, :, :]

        assert_allclose(integrals.d_distance(R, mesh.face_normals), z_offset)


def test_dipole_potential_approximation():
    """
    Test whether approximations and exact solution are close enough

    """

    mesh = load_example_mesh("unit_disc")

    points = np.array([[0.1, 0, 0.2], [-0.1, 0.1, -0.2]]) * 20

    R = points[:, None, None, :] - mesh.vertices[mesh.faces][None, :, :, :]
    R_v = points[:, None, :] - mesh.vertices[None, :, :]

    vertex_areas = mass_matrix(mesh, lumped=True).diagonal()

    approx_pot_v = integrals.potential_vertex_dipoles(
        R_v, mesh.vertex_normals, vertex_areas
    )

    approx_pot_f = integrals.potential_dipoles(R, mesh.face_normals, mesh.area_faces)
    exact_pot_f = integrals.triangle_potential_dipole_linear(
        R, mesh.face_normals, mesh.area_faces
    )

    # Map faces -> vertices
    Nf = len(mesh.faces)
    Nv = len(mesh.vertices)
    M0 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 0])), (Nf, Nv))
    M1 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 1])), (Nf, Nv))
    M2 = csc_matrix((np.ones(Nf), (np.arange(Nf), mesh.faces[:, 2])), (Nf, Nv))
    exact_pot_v = (
        exact_pot_f[:, :, 0] @ M0
        + exact_pot_f[:, :, 1] @ M1
        + exact_pot_f[:, :, 2] @ M2
    )
    approx_pot_fv = (
        approx_pot_f[:, :, 0] @ M0
        + approx_pot_f[:, :, 1] @ M1
        + approx_pot_f[:, :, 2] @ M2
    )

    assert_allclose(approx_pot_v, exact_pot_v, rtol=5e-2)
    assert_allclose(approx_pot_fv, exact_pot_v, rtol=1e-3)


def test_triangle_potential_approximation():
    """
    test whether 1/r triangle potential approximations give close enough
    to exact (with/without planar assumption)
    """

    mesh = load_example_mesh("unit_disc")

    points = np.array([[0, 0, 1], [-0.1, 0.1, -5]])

    R = points[:, None, None, :] - mesh.vertices[mesh.faces][None, :, :, :]
    Rcenters = points[:, None, :] - mesh.triangles_center[None, :, :]

    exact = integrals.triangle_potential_uniform(R, mesh.face_normals)
    approx = integrals.triangle_potential_approx(Rcenters, mesh.area_faces)

    assert_allclose(approx, exact, rtol=5e-3)

    points = np.array([[0, 0, 0], [-0.1, 0.1, 0]])
    R = points[:, None, None, :] - mesh.vertices[mesh.faces][None, :, :, :]

    exact = integrals.triangle_potential_uniform(R, mesh.face_normals)
    exact_planar = integrals.triangle_potential_uniform(
        R, mesh.face_normals, planar=True
    )

    assert_allclose(exact_planar, exact)



"""
Analytic B-field computation
==================================================
Validation of analytic mesh operator for magnetic field computation.
"""

import numpy as np
import trimesh
from mayavi import mlab
import matplotlib.pyplot as plt

from bfieldtools.mesh_calculus import gradient
from bfieldtools.mesh_magnetics import (
    magnetic_field_coupling,
    magnetic_field_coupling_analytic,
)
from bfieldtools.mesh_conductor import MeshConductor
import pkg_resources


# Load simple plane mesh that is centered on the origin
file_obj = pkg_resources.resource_filename(
    "bfieldtools", "example_meshes/10x10_plane.obj"
)
coilmesh = trimesh.load(file_obj, process=False)
coil = MeshConductor(mesh_obj=coilmesh)
weights = np.zeros(coilmesh.vertices.shape[0])
weights[coil.inner_vertices] = 1

test_points = coilmesh.vertices + np.array([0, 1, 0])

B0 = magnetic_field_coupling(coilmesh, test_points) @ weights
B1 = magnetic_field_coupling_analytic(coilmesh, test_points) @ weights


f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

s = mlab.triangular_mesh(
    *coilmesh.vertices.T, coilmesh.faces, scalars=weights, colormap="viridis"
)
s.enable_contours = True
s.actor.property.render_lines_as_tubes = True
s.actor.property.line_width = 3.0

mlab.quiver3d(*test_points.T, *B0.T, color=(1, 0, 0))
mlab.quiver3d(*test_points.T, *B1.T, color=(0, 0, 1))

print(
    "Relative RMS error", np.sqrt(np.mean((B1 - B0) ** 2)) / np.sqrt(np.mean((B0) ** 2))
)

#%% Test against analytic formula
# Load simple plane mesh that is centered on the origin
file_obj = pkg_resources.resource_filename(
    "bfieldtools", "example_meshes/unit_disc.stl"
)
discmesh = trimesh.load(file_obj, process=True)
for ii in range(3):
    discmesh = discmesh.subdivide()
disc = MeshConductor(mesh_obj=discmesh)
weights = np.zeros(discmesh.vertices.shape[0])
weights[disc.inner_vertices] = 1
mlab.figure()
s = mlab.triangular_mesh(
    *discmesh.vertices.T, discmesh.faces, scalars=weights, colormap="viridis"
)
g = gradient(weights, discmesh, rotated=True)
mlab.quiver3d(*discmesh.vertices[discmesh.faces].mean(axis=1).T, *g)

test_points = np.zeros((100, 3))
test_points[:, 2] = np.linspace(0.0, 5, 100)
mlab.points3d(*test_points.T, scale_factor=0.1)

# Bfield for 1 Ampere current
B0 = magnetic_field_coupling(discmesh, test_points) @ weights
B1 = magnetic_field_coupling_analytic(discmesh, test_points) @ weights

# Analytic formula for unit disc
plt.plot(1e-7 * 2 * np.pi / (np.sqrt(test_points[:, 2] ** 2 + 1) ** 3))
# Field from the mesh
plt.plot(np.linalg.norm(B0, axis=1))
plt.plot(np.linalg.norm(B1, axis=1))

plt.legend(("Analytic", "Quadrature mesh", "Analytic mesh"))
plt.xlabel("Distance, z [m]")
plt.ylabel("B [T]")



# -*- coding: utf-8 -*-
"""
Validation of Laplacian using spherical harmonics
====================================================================

Study the eigenvalue spectrum of the discretize laplace-beltrami operator
on a spherical mesh. Compare the spectrum to analytical solution.
"""

import numpy as np
from bfieldtools.mesh_calculus import mass_matrix, laplacian_matrix
from bfieldtools.utils import load_example_mesh
from scipy.sparse.linalg import eigsh
import matplotlib.pyplot as plt
import trimesh

# This is icosphere(4)?
# mesh = load_example_mesh("unit_sphere")
# The test is faster with a smaller number of vertices
mesh = trimesh.creation.icosphere(3)

Nvals = 150
uu1 = []
vv1 = []
M = mass_matrix(mesh)

# for q in range(4):
L = laplacian_matrix(mesh)
uu, vv = eigsh(-L, Nvals, M, which="SM")
uu1.append(uu)
vv1.append(vv)


#%%
""" 
 Spherical harmonics are the eigenfunctions of the LB operator
 The correct eigenvalues are l*(l+1)
"""
R = np.linalg.norm(mesh.vertices[mesh.faces].mean(axis=1), axis=-1).mean()
ll = np.array([l for l in range(20) for m in range(-l, l + 1)])
ll = ll[:Nvals]
evals = ll * (ll + 1)

#%% Plot relateive errors in eigenvalues
# plt.plot(evals, 'k')
for u in uu1:
    plt.plot(abs(evals[1:] - u[1:]) / evals[1:], "-")

plt.xlabel("# eigenvalue")
plt.ylabel("Relative error")

#%% Eigenfunctions
from bfieldtools.utils import MeshProjection
from bfieldtools.sphtools import ylm
from bfieldtools.sphtools import cartesian2spherical

ylm_on_hats = []
i1 = 0
vv1_projs = np.zeros((len(vv1), vv1[0].shape[1]))
mp = MeshProjection(mesh, 4)

# N < L*(L+2)
for l in range(0, 13):
    i0 = i1
    print(f"l={l}")
    for m in range(-l, l + 1):

        def func(r):
            sphcoords = cartesian2spherical(r)
            return ylm(l, m, sphcoords[:, 1], sphcoords[:, 2])

        ylm_on_hats.append(mp.hatfunc_innerproducts(func))
        i1 += 1
    for ii, vv in enumerate(vv1):
        # Project self-inductance eigenfunctions to l-subspace
        p = np.sum((np.array(ylm_on_hats[i0:i1]) @ vv[:, i0:i1]) ** 2, axis=0)
        vv1_projs[ii, i0:i1] = p

#%%
"""
Plot the norm of a projection of the eigenfunctions into the L-subspaces
corresponding to the same degenerate eigenvalue.
The correct values should be 1, if they are less than one the eigenfunctions
span also to other subspaces
"""
plt.figure()
eff_R2 = np.sum(mesh.area_faces) / (4 * np.pi)
plt.plot(np.sqrt(vv1_projs.T / eff_R2), "-")

plt.xlabel("# eigenfunction")
plt.ylabel("Squared norm in L-subspace")



"""
Module with functions for exporting polylines to the PCB design software KiCAD
"""

__all__ = ["python_to_kicad"]


def python_to_kicad(
    loops, filename, plane_axes, origin, layer, net, scaling=1, trace_width=0.2
):
    """
    Appends polylines to KiCAD PCB files as traces.

    Paramaters
    ----------
        loops: N_loops long list of (Nv_loop, 3) arrays
            Each corresponding to the vertex locations for widning loops
        filename: str
            filename/path where the ifno is written (append mode is used)
        plane_axes: tuple with length 2
            specifies the X- and Y-dimensions used for the PCB
        origin: (2, ) array-like
            Origin shift applid in original units
        layer: str
            Which layer to write to
        scaling: float
            Scaling factor applied to the loops
        trace_width: float
            The trace width in mm
        net: int
            Specifies which net number the loops are assigned to

    Returns
    -------
        None

    """

    with open(filename, "a") as file:
        for loop in loops:
            for seg_idx in range(1, len(loop)):
                x_start = loop[seg_idx - 1, plane_axes[0]] + origin[0]
                y_start = loop[seg_idx - 1, plane_axes[1]] + origin[1]

                x_end = loop[seg_idx, plane_axes[0]] + origin[0]
                y_end = loop[seg_idx, plane_axes[1]] + origin[1]

                file.write(
                    "    (segment (start %.2f %.4f) (end %.2f %.2f) (width %.2f) (layer %s) (net %d))\n"
                    % (
                        x_start * scaling,
                        y_start * scaling,
                        x_end * scaling,
                        y_end * scaling,
                        trace_width,
                        layer,
                        net,
                    )
                )

    return



"""
Analytical self-shielded biplanar coil design
==============================================

Example showing a basic biplanar coil producing homogeneous field in a target
region between the two coil planes. In addition, the coils have an outer surface
for which (in a linear fashion) a secondary current is created, which zeroes the
normal component of the field produced by the primary coil at the secondary coil
surface. The combination of the primary and secondary coil currents are specified to create
the target field, and their combined inductive energy is minimized.

NB. The secondary coil current is entirely a function of the primary coil current
and the geometry.
"""

import numpy as np
import matplotlib.pyplot as plt
from mayavi import mlab
import trimesh

from bfieldtools.mesh_conductor import MeshConductor, StreamFunction
from bfieldtools.utils import combine_meshes, load_example_mesh

# Load simple plane mesh that is centered on the origin
planemesh = load_example_mesh("10x10_plane_hires")


# Specify coil plane geometry
center_offset = np.array([0, 0, 0])
standoff = np.array([0, 4, 0])

# Create coil plane pairs
coil_plus = trimesh.Trimesh(
    planemesh.vertices + center_offset + standoff, planemesh.faces, process=False
)

coil_minus = trimesh.Trimesh(
    planemesh.vertices + center_offset - standoff, planemesh.faces, process=False
)

mesh1 = combine_meshes((coil_plus, coil_minus))
mesh2 = mesh1.copy()
mesh2.apply_scale(1.4)

coil = MeshConductor(mesh_obj=mesh1, basis_name="inner", N_sph=4)
shieldcoil = MeshConductor(mesh_obj=mesh2, basis_name="inner", N_sph=4)

#%%
# Plot geometry
f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))
coil.plot_mesh(opacity=0.2, figure=f)
shieldcoil.plot_mesh(opacity=0.2, figure=f)

#%%
# Compute inductances and coupling


M11 = coil.inductance
M22 = shieldcoil.inductance
M21 = shieldcoil.mutual_inductance(coil)


# Mapping from I1 to I2, constraining flux through shieldcoil to zero
P = -np.linalg.solve(M22, M21)

A1, Beta1 = coil.sph_couplings
A2, Beta2 = shieldcoil.sph_couplings

#%%
# Precalculations for the solution

# Minimization of magnetic energy with spherical harmonic constraint
C = Beta1 + Beta2 @ P
M = M11 + M21.T @ P

# Regularization
from scipy.linalg import eigvalsh

ssmax = eigvalsh(C.T @ C, M, eigvals=[M.shape[1] - 1, M.shape[1] - 1])

#%%
# Specify spherical harmonic and calculate corresponding shielded field

beta = np.zeros(Beta1.shape[0])
# beta[7] = 1 # Gradient
beta[2] = 1  # Homogeneous

# Minimum residual
_lambda = 1e3
# Minimum energy
# _lambda=1e-3
I1inner = np.linalg.solve(C.T @ C + M * ssmax / _lambda, C.T @ beta)

I2inner = P @ I1inner

coil.s = StreamFunction(I1inner, coil)
shieldcoil.s = StreamFunction(I2inner, shieldcoil)

#%%
# Do a quick 3D plot

f = mlab.figure(None, bgcolor=(1, 1, 1), fgcolor=(0.5, 0.5, 0.5), size=(800, 800))

coil.s.plot(figure=f, contours=20)
shieldcoil.s.plot(figure=f, contours=20)

#%%
# Compute the field and scalar potential on an XY-plane

x = y = np.linspace(-8, 8, 150)
X, Y = np.meshgrid(x, y, indexing="ij")
points = np.zeros((X.flatten().shape[0], 3))
points[:, 0] = X.flatten()
points[:, 1] = Y.flatten()


CB1 = coil.B_coupling(points)
CB2 = shieldcoil.B_coupling(points)

CU1 = coil.U_coupling(points)
CU2 = shieldcoil.U_coupling(points)

B1 = CB1 @ coil.s
B2 = CB2 @ shieldcoil.s

U1 = CU1 @ coil.s
U2 = CU2 @ shieldcoil.s


#%%
# Now, plot the field streamlines and scalar potential


from bfieldtools.contour import scalar_contour

cc1 = scalar_contour(mesh1, mesh1.vertices[:, 2], contours=[-0.001])
cc2 = scalar_contour(mesh2, mesh2.vertices[:, 2], contours=[-0.001])
cx10 = cc1[0][:, 1]
cy10 = cc1[0][:, 0]
cx20 = cc2[0][:, 1]
cy20 = cc2[0][:, 0]

cx11 = np.vstack(cc1[1:])[:, 1]
cy11 = np.vstack(cc1[1:])[:, 0]
cx21 = np.vstack(cc2[1:])[:, 1]
cy21 = np.vstack(cc2[1:])[:, 0]

B = (B1.T + B2.T)[:2].reshape(2, x.shape[0], y.shape[0])
lw = np.sqrt(B[0] ** 2 + B[1] ** 2)
lw = 2 * np.log(lw / np.max(lw) * np.e + 1.1)

xx = np.linspace(-1, 1, 16)

seed_points = np.array([cx10 + 0.001, cy10])
seed_points = np.hstack([seed_points, np.array([cx11 - 0.001, cy11])])
seed_points = np.hstack([seed_points, (0.56 * np.array([np.zeros_like(xx), xx]))])


U = (U1 + U2).reshape(x.shape[0], y.shape[0])
U /= np.max(U)
plt.figure()
plt.contourf(X, Y, U.T, cmap="seismic", levels=40)
# plt.imshow(U, vmin=-1.0, vmax=1.0, cmap='seismic', interpolation='bicubic',
#           extent=(x.min(), x.max(), y.min(), y.max()))
plt.streamplot(
    x,
    y,
    B[1],
    B[0],
    density=2,
    linewidth=lw,
    color="k",
    start_points=seed_points.T,
    integration_direction="both",
    arrowsize=0.1,
)

for loop in cc1 + cc2:
    plt.plot(loop[:, 1], loop[:, 0], "k", linewidth=4, alpha=1)
    plt.plot(-loop[:, 1], -loop[:, 0], "k", linewidth=4, alpha=1)

plt.axis("image")

plt.xticks([])
plt.yticks([])



from setuptools import setup, find_packages
import sys
import codecs


def main():

    with open("requirements.txt") as f:
        requirements = f.read().splitlines()

    setup(
        maintainer="Rasmus Zetter",
        maintainer_email="rasmus.zetter@aalto.fi",
        name="bfieldtools",
        description="Open-source Python software for magnetic field modeling",
        long_description=codecs.open("README.rst", encoding="utf8").read(),
        url="https://bfieldtools.github.io/",
        license="BSD (3-clause)",
        download_url="https://github.com/bfieldtools/bfieldtools",
        use_scm_version={
            "write_to": "bfieldtools/version.py",
            "write_to_template": '__version__ = "{version}"',
        },
        setup_requires=["setuptools_scm"],
        python_requires=">=3.6",
        install_requires=requirements,
        packages=find_packages(),
        include_package_data=True,
        package_data={"bfieldtools": ["example_meshes/*"]},
        classifiers=[
            "Intended Audience :: Science/Research",
            "Intended Audience :: Developers",
            "Programming Language :: Python",
            "Topic :: Software Development",
            "Topic :: Scientific/Engineering",
        ],
        platforms="any",
        zip_safe=False,
    )


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 17 16:58:55 2020

@author: Rasmus Zetter
"""

"""
Contains functions for computing vector calculus quantities on triangle surface meshes.
These include the gradient, rotated gradient, divergence, curl, Laplacian and mass matrices.
"""

import numpy as np
from scipy.sparse import csr_matrix, coo_matrix, spdiags, hstack, vstack


def laplacian_matrix(mesh, material_param=None, inner_vertices=None, holes=None):
    """
    Sparse Laplace-Beltrami operator

    If inner vertices are not given Laplacian for all the mesh vertices is returned.
    This corresponds to zero-Neumann (natural) boundary condition on the possible
    outer boundary.
    shape==(len(mesh.vertices), len(mesh.vertices))

    If inner_vertices but no holes are given, the outer boundary is assumed grounded
    and Laplacian only for the inner vertices is returned
    shape==(len(inner_vertices), len(inner_vertices))

    If both inner_vertices and holes are given, the outer boundary is assumed grounded
    and constant floating boundary condition for each hole is assumed
    shape==(len(inner_vertices)+len(holes), len(inner_vertices)+len(holes))

    Parameters
    ----------
    mesh: Trimesh Mesh object
    material_param: array-like with length N_triangles
        material parameter for each triangle
    inner_vertices: list (default None)
        contains mesh vertex indices corresponding to inner holes
    holes: list with length N_holes (default None)
        each list element contains array-like of mesh vertex indices corresponding to each
        mesh hole

    Returns
    -------
    sparse csr_matrix of variable shape (see description)
        cotangent-laplacian: w_ij = - 0.5* (cot(alpha) + cot(beta))

    """
    if material_param is None:
        p = 1
    else:
        p = material_param
    N = mesh.vertices.shape[0]
    R = mesh.vertices[mesh.faces]  # Nt x 3 (corners) x 3 (xyz)

    # Edges opposite to the vertex
    edges = np.roll(R, 1, -2) - np.roll(R, 2, -2)  # Nt x 3 (edges) x 3 (x,y,z)

    ii = []
    jj = []
    cot = []
    # Loop over edges in triangles
    for i in range(3):
        i1 = (i + 1) % 3
        i2 = (i + 2) % 3
        ii.append(mesh.faces[:, i1])
        jj.append(mesh.faces[:, i2])
        c = (
            -0.5
            * (edges[:, i1, :] * edges[:, i2, :]).sum(axis=-1)
            / (2 * mesh.area_faces)
        )
        # Append cot with cotangent terms multiplied by material parameter
        cot.append(c * p)

    ii = np.ravel(ii)
    jj = np.ravel(jj)
    cot = np.ravel(cot)
    # Build sparse matrix
    L = csr_matrix((cot, (ii, jj)), shape=(N, N), dtype=float)
    # Sum contribution from both triangles (alpha and beta angles)
    # neighbouring the edge
    L = L + L.T
    L = L - spdiags(L.sum(axis=0), 0, N, N)

    # If inner_vertices are specified, return matrix only for those vertices
    if inner_vertices is not None:
        # If holes specified, add matrix entries corresponding to them
        if holes:
            L = _laplacian_matrix_w_holes(L, inner_vertices, holes)
        else:
            L = L[inner_vertices][:, inner_vertices]
    # Catch if only holes specified, but not inner_vertices
    elif holes:
        raise ValueError("You need to specify both inner_vertices and holes")

    return L


def _laplacian_matrix_w_holes(L, inner_vertices, holes):
    """
    Computes Laplacian matrix with additional boundary constraint
    for inner boundaris: the transverse gradient at the inner holes is zero,
    i.e. the value on the hole boundary is constant
    Mesh vertices not present in inner_vertices or holes are assumed to be
    on the outer boundary of the mesh, which is set to zero.

    For the discretization see:
    https://www.cs.cmu.edu/~kmcrane/Projects/Other/SwissArmyLaplacian.pdf
    page 232

    When the values on the boundary are constrained equal, we have only one
    degree of freedom and the element corresponding to that corresponds
    to summing the individual elements in the original matrix


    Parameters
    ----------
    L: Laplacian matrix computes without boundary conditions

    inner_vertices: list
        contains mesh vertex indices corresponding to inner holes
    holes: list with length N_holes
        each list element contains array-like of mesh vertex indices corresponding to each
        mesh hole

    Returns
    -------
    sparse csr_matrix
        Laplacian modified for holes
        first N_inner_vertices elements correspond to inner mesh vertices,
        last N_holes elements correspond to the values at the holes

    """

    Lb = [None] * len(holes)

    # Start constructing the Laplacian matrix including the values at the inner holes
    L_holes = L[inner_vertices, :][:, inner_vertices]

    # Add columns
    for b_idx, b in enumerate(holes):
        # Hole contribution in original Laplacian matrix
        Lb[b_idx] = coo_matrix(np.sum(L[b, :][:, inner_vertices], axis=0))

        # Add on the values at the right-hand side of the matrix
        L_holes = hstack((L_holes, Lb[b_idx].T))

    # Add rows, including new diagonal
    for b_idx, b in enumerate(holes):
        # Construct the added-on diagonal part
        concat = np.zeros((len(holes), 1))
        concat[b_idx] = -np.sum(Lb[b_idx])

        # Add on the values at the bottom of the matrix, including the diagonal part
        L_holes = vstack((L_holes, hstack((Lb[b_idx], coo_matrix(concat.T)))))

    return L_holes.tocsr()


def mass_matrix(mesh, lumped=False, inner_vertices=None, holes=None):
    """
    Computes mass matrix of mesh.

    Parameters
    ----------
    mesh: Trimesh Mesh object
    lumped: Boolean
        If True, use lumped approximation of mass matrix. If False (default),
        compute exact matrix. See Reuter et al 2009, page 3 (DOI: 10.1016/j.cag.2009.03.005)
    inner_vertices: list (default None)
        contains mesh vertex indices corresponding to inner holes
    holes: list with length N_holes (default None)
        each list element contains array-like of mesh vertex indices corresponding to each
        mesh hole

    Returns
    -------
    sparse csr_matrix of variable shape (see the description of laplacian_matrix)
        Mass matrix

    """

    if lumped:
        from .utils import dual_areas

        da = dual_areas(mesh.faces, mesh.area_faces)
        M = spdiags(da, 0, mesh.vertices.shape[0], mesh.vertices.shape[0]).tocsr()
    else:
        N = mesh.vertices.shape[0]

        ii = []
        jj = []
        area = []
        # Loop over edges in triangles
        for i in range(3):
            i1 = (i + 1) % 3
            i2 = (i + 2) % 3
            ii.append(mesh.faces[:, i1])
            jj.append(mesh.faces[:, i2])
            area.append(mesh.area_faces / 12)

        ii = np.ravel(ii)
        jj = np.ravel(jj)
        area = np.ravel(area)
        # Build sparse matrix
        M = csr_matrix((area, (ii, jj)), shape=(N, N), dtype=float)
        # Sum contribution from both triangles (alpha and beta angles)
        # neighbouring the edge
        M = M + M.T
        M = M + spdiags(M.sum(axis=0), 0, N, N)

    # If inner_vertices specified, return only values for inner_vertices
    if inner_vertices is not None:
        # If holes are specifed, add corresponding values
        if holes:
            M = _mass_matrix_w_holes(M, inner_vertices, holes)
        else:
            M = M[inner_vertices, :][:, inner_vertices]
    elif holes:
        raise ValueError("You need to specify both inner_vertices and holes")

    return M


def _mass_matrix_w_holes(M, inner_vertices, holes):
    """
    Computes mass matrix of mesh with added holes (see laplacian_matrix_w_holes)
    """

    Minner = M[inner_vertices, :][:, inner_vertices]
    m = M.diagonal()

    M_holes = Minner.diagonal()
    for b in holes:
        M_holes = np.concatenate((M_holes, np.array([np.sum(m[b])])))

    M_holes = spdiags(
        M_holes, diags=0, m=M_holes.shape[0], n=M_holes.shape[0], format="csr"
    )

    return M_holes
